from typing import List, Dict, Optional, Tuple
import pandas as pd
from datasets import load_dataset
from os.path import join
from datasets import DatasetDict

from nerblackbox.modules.datasets.formatter.base_formatter import (
    BaseFormatter,
    SENTENCES_ROWS,
)

SUCX_SUBSETS = [
    "original_cased",
    "original_lower",
    "original_lower_mix",
    "simple_cased",
    "simple_lower",
    "simple_lower_mix",
]
SUCX_NER_TAG_LIST = {
    "original": [
        "person",
        "place",
        "inst",
        "work",
        "product",
        "animal",
        "event",
        "myth",
        "other",
    ],
    "simple": ["PER", "LOC", "ORG", "WRK", "OBJ", "PRS", "EVN", "PRS", "MSR", "TME"],
}


class SUCXFormatter(BaseFormatter):
    def __init__(self, ner_dataset_subset: str):
        ner_dataset = "sucx"
        assert (
            ner_dataset_subset in SUCX_SUBSETS
        ), f"ERROR! subset = {ner_dataset_subset} unknown."

        tag_group = ner_dataset_subset.split("_")[0]
        assert tag_group in [
            "original",
            "simple",
        ], f"ERROR! tag_group = {tag_group} should be original or simple."
        ner_tag_list = SUCX_NER_TAG_LIST[tag_group]

        super().__init__(ner_dataset, ner_tag_list, ner_dataset_subset)
        self.ner_dataset_subset = ner_dataset_subset

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool) -> None:  # pragma: no cover
        """
        I: get data

        Args:
            verbose: [bool]
        """
        dataset = load_dataset("KBLab/sucx3_ner", self.ner_dataset_subset)
        assert isinstance(
            dataset, DatasetDict
        ), f"ERROR! type(dataset) = {type(dataset)} expected to be DatasetDict."
        dataset["val"] = dataset["validation"]
        for phase in ["train", "val", "test"]:
            sentences_rows_formatted = []
            for sample in dataset[phase]:
                sentences_rows_formatted.append(
                    (" ".join(sample["ner_tags"]), " ".join(sample["tokens"]))
                )

            df = pd.DataFrame(sentences_rows_formatted)
            assert (
                len(df) == dataset[phase].num_rows
            ), f"ERROR! number of rows = {len(df)} != {dataset[phase].num_rows}"
            file_path = join(self.dataset_path, f"{phase}_original.csv")
            df.to_csv(file_path, sep="\t", header=False, index=False)

    def create_ner_tag_mapping(self) -> Dict[str, str]:
        """
        II: customize ner_training tag mapping if wanted

        Returns:
            ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return dict()

    def format_data(
        self, shuffle: bool = True, write_csv: bool = True
    ) -> Optional[SENTENCES_ROWS]:
        """
        III: format data

        Args:
            shuffle: whether to shuffle rows of dataset
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            sentences_rows: only if write_csv = False
        """
        self.set_original_file_paths()
        for phase in ["train", "val", "test"]:
            file_path_original = join(self.dataset_path, self.file_name[phase])
            df = pd.read_csv(file_path_original, sep="\t", header=None)
            _sentences_rows = [_np_array for _np_array in df.values]
            sentences_rows = [
                [[word, tag] for word, tag in zip(words.split(), tags.split())]
                for tags, words in _sentences_rows
            ]
            if shuffle:
                sentences_rows = self._shuffle_dataset(phase, sentences_rows)

            if write_csv:  # pragma: no cover
                self._write_formatted_csv(phase, sentences_rows)
            else:
                return sentences_rows
        return None

    def set_original_file_paths(self) -> None:
        """
        III: format data

        Changed Attributes:
            file_paths: [Dict[str, str]], e.g. {'train': <path_to_train_csv>, 'val': ..}

        Returns: -
        """
        self.file_name = {
            "train": "train_original.csv",
            "val": "val_original.csv",
            "test": "test_original.csv",
        }

    def _parse_row(self, _row: str) -> List[str]:
        """
        III: format data

        Args:
            _row: e.g. "Det PER X B"

        Returns:
            _row_list: e.g. ["Det", "PER", "X", "B"]
        """
        return []

    def _format_original_file(self, _row_list: List[str]) -> Optional[List[str]]:
        """
        III: format data

        Args:
            _row_list: e.g. ["B-product I-product I-product B-product	Audi Coupé Quattro 20V"]

        Returns:
            _row_list_formatted: e.g. ["test", "B-PER"]
        """
        pass

    def resplit_data(
        self, val_fraction: float = 0.0, write_csv: bool = True
    ) -> Optional[Tuple[pd.DataFrame, ...]]:
        """
        IV: resplit data

        Args:
            val_fraction: [float], e.g. 0.3
            write_csv: whether to write dataset to csv (should always be True except for testing)

        Returns:
            df_train: only if write_csv = False
            df_val:   only if write_csv = False
            df_test:  only if write_csv = False
        """
        # train -> train
        df_train = self._read_formatted_csvs(["train"])

        # val  -> val
        df_val = self._read_formatted_csvs(["val"])

        # test  -> test
        df_test = self._read_formatted_csvs(["test"])

        if write_csv:  # pragma: no cover
            self._write_final_csv("train", df_train)
            self._write_final_csv("val", df_val)
            self._write_final_csv("test", df_test)
            return None
        else:
            return df_train, df_val, df_test



import pytest
from nerblackbox.modules.datasets.formatter.base_formatter import SENTENCES_ROWS
from nerblackbox.modules.datasets.formatter.sucx_formatter import SUCXFormatter
from pkg_resources import resource_filename
import pandas as pd

import os
from os.path import abspath, dirname, join

BASE_DIR = abspath(dirname(dirname(dirname(__file__))))
DATA_DIR = join(BASE_DIR, "data")
os.environ["DATA_DIR"] = DATA_DIR


class TestSucxFormatter:

    formatter = SUCXFormatter("original_cased")

    @pytest.mark.parametrize(
        "sentences_rows",
        [
            (
                [
                    [
                        ["I", "O"],
                        ["850-modellen", "O"],
                        ["har", "O"],
                        ["man", "O"],
                        ["valt", "O"],
                        ["en", "O"],
                        ["tredje", "O"],
                        ["variant", "O"],
                        [",", "O"],
                        ["Delta-link", "B-product"],
                        [".", "O"],
                    ],
                    [
                        ["Audi", "B-product"],
                        ["Coupé", "I-product"],
                        ["Quattro", "I-product"],
                        ["20V", "B-product"],
                    ],
                ]
            ),
        ],
    )
    def test_format_data(self, sentences_rows: SENTENCES_ROWS):
        self.formatter.dataset_path = resource_filename(
            "nerblackbox", f"tests/test_data/original_data"
        )
        test_sentences_rows = self.formatter.format_data(shuffle=True, write_csv=False)
        assert (
            test_sentences_rows == sentences_rows
        ), f"ERROR! test_sentences_rows = {test_sentences_rows} != {sentences_rows}"

    @pytest.mark.parametrize(
        "df_train, df_val, df_test",
        [
            (
                [
                    pd.DataFrame(
                        data=[
                            ["O O", "Mening 1"],
                            ["PER O", "Mening 2"],
                            ["O PER", "Mening 3"],
                            ["O PER", "Mening 4"],
                        ]
                    ),
                    pd.DataFrame(data=[["O O", "Mening 5"], ["PER O", "Mening 6"]]),
                    pd.DataFrame(data=[["O O", "Mening 7"], ["PER O", "Mening 8"]]),
                ]
            ),
        ],
    )
    def test_resplit_data(
        self, df_train: pd.DataFrame, df_val: pd.DataFrame, df_test: pd.DataFrame
    ):
        self.formatter.dataset_path = resource_filename(
            "nerblackbox", f"tests/test_data/formatted_data"
        )
        test_dfs = self.formatter.resplit_data(write_csv=False)
        assert (
            test_dfs is not None
        ), f"ERROR! resplit_data() returned None unexpectedly."
        test_df_train, test_df_val, test_df_test = test_dfs
        pd.testing.assert_frame_equal(
            test_df_train, df_train
        ), f"ERROR! test_resplit_data did not pass test for phase = train"
        pd.testing.assert_frame_equal(
            test_df_val, df_val
        ), f"ERROR! test_resplit_data did not pass test for phase = val"
        pd.testing.assert_frame_equal(
            test_df_test, df_test
        ), f"ERROR! test_resplit_data did not pass test for phase = test"



import pytest
from typing import Dict, List, Any, Tuple, Union
import numpy as np

import torch

from nerblackbox.api.model import (
    EVALUATION_DICT,
    round_evaluation_dict,
    derive_annotation_scheme,
    turn_tensors_into_tag_probability_distributions,
    merge_slices_for_single_document,
    merge_subtoken_to_token_predictions,
    restore_unknown_tokens,
    assert_typing,
)


class TestModelStatic:
    @pytest.mark.parametrize(
        "evaluation_dict, rounded_decimals, evaluation_dict_rounded",
        [
            (
                {
                    "micro": {
                        "entity": {
                            "precision": 0.9847222222222223,
                            "recall": 0.9916083916083916,
                            "f1": 0.9881533101045297,
                            "precision_seqeval": 0.9833564493758669,
                            "recall_seqeval": 0.9916083916083916,
                            "f1_seqeval": 0.9874651810584958,
                        },
                    }
                },
                3,
                {
                    "micro": {
                        "entity": {
                            "precision": 0.985,
                            "recall": 0.992,
                            "f1": 0.988,
                            "precision_seqeval": 0.983,
                            "recall_seqeval": 0.992,
                            "f1_seqeval": 0.987,
                        }
                    }
                },
            ),
        ],
    )
    def test_derive_evaluation_dict(
        self,
        evaluation_dict: EVALUATION_DICT,
        rounded_decimals: int,
        evaluation_dict_rounded: EVALUATION_DICT,
    ):
        test_evaluation_dict_rounded = round_evaluation_dict(
            evaluation_dict, rounded_decimals
        )
        assert test_evaluation_dict_rounded == evaluation_dict_rounded, (
            f"ERROR! test_evaluation_dict_rounded = {test_evaluation_dict_rounded} "
            f"!= {evaluation_dict_rounded} = evaluation_dict_rounded"
        )

    @pytest.mark.parametrize(
        "id2label, error, annotation_scheme",
        [
            (
                {0: "O", 1: "PER"},
                False,
                "plain",
            ),
            (
                {0: "O", 1: "B-PER", 2: "I-PER"},
                False,
                "bio",
            ),
            (
                {0: "O", 1: "B-PER", 2: "I-PER", 3: "L-PER", 4: "U-PER"},
                False,
                "bilou",
            ),
            (
                {0: "O", 1: "A-PER", 2: "B-PER"},
                True,
                "",
            ),
            (
                {},
                True,
                "",
            ),
        ],
    )
    def test_derive_annotation_scheme(
        self, id2label: Dict[int, str], error: bool, annotation_scheme: str
    ):
        if error:
            with pytest.raises(Exception):
                _ = derive_annotation_scheme(id2label)
        else:
            test_annotation_scheme = derive_annotation_scheme(id2label)
            assert (
                test_annotation_scheme == annotation_scheme
            ), f"ERROR! test_annotation_scheme = {test_annotation_scheme} != {annotation_scheme} = annotation_scheme"

    @pytest.mark.parametrize(
        "annotation_classes, outputs, predictions",
        [
            (
                ["O", "B-PER", "I-PER"],
                torch.tensor([[[0.01, 3, 0.01]]]),
                [[{"O": 0.0457, "B-PER": 0.9086, "I-PER": 0.0457}]],
            ),
        ],
    )
    def test_turn_tensors_into_tag_probability_distributions(
        self,
        annotation_classes: List[str],
        outputs: torch.Tensor,
        predictions: List[List[Dict[str, float]]],
    ):
        test_predictions = turn_tensors_into_tag_probability_distributions(
            annotation_classes, outputs
        )
        for list1, list2 in zip(test_predictions, predictions):
            for dict1, dict2 in zip(list1, list2):
                assert set(dict1.keys()) == set(
                    dict2.keys()
                ), f"ERROR! test keys = {set(dict1.keys())} != {set(dict2.keys())} = keys"
                for k in dict1:
                    assert (
                        np.absolute(dict1[k] - dict2[k]) < 0.0001
                    ), f"ERROR! test_predictions = {dict1} != {dict2} = predictions"

    @pytest.mark.parametrize(
        "list_slices, list_documents",
        [
            # 1. tokens
            (
                [["[CLS]", "this", "is", "one", "slice", "[SEP]"]],
                ["[CLS]", "this", "is", "one", "slice", "[SEP]"],
            ),
            (
                [
                    ["[CLS]", "this", "is", "one", "slice", "[SEP]"],
                    ["[CLS]", "and", "a", "second", "one", "[SEP]"],
                ],
                [
                    "[CLS]",
                    "this",
                    "is",
                    "one",
                    "slice",
                    "and",
                    "a",
                    "second",
                    "one",
                    "[SEP]",
                ],
            ),
            (
                [
                    ["[CLS]", "slice", "1", "[SEP]"],
                    ["[CLS]", "slice", "2", "[SEP]"],
                    ["[CLS]", "slice", "3", "[SEP]"],
                ],
                ["[CLS]", "slice", "1", "slice", "2", "slice", "3", "[SEP]"],
            ),
            # 2. predictions
            (
                [
                    ["O", "B-PER", "I-PER", "[S]"],
                    ["[S]", "B-LOC", "I-LOC", "[SEP]"],
                    ["[S]", "B-ORG", "I-ORG", "O"],
                ],
                ["O", "B-PER", "I-PER", "B-LOC", "I-LOC", "B-ORG", "I-ORG", "O"],
            ),
            # 3. predictions with proba = True
            (
                [
                    [
                        {"O": 0.1, "PER": 0.9},
                        {"O": 0.1, "PER": 0.9},
                        {"O": 1.1, "PER": 1.1},
                    ],
                    [
                        {"O": 1.1, "PER": 1.1},
                        {"O": 0.1, "PER": 0.9},
                        {"O": 0.1, "PER": 0.9},
                    ],
                ],
                [
                    {"O": 0.1, "PER": 0.9},
                    {"O": 0.1, "PER": 0.9},
                    {"O": 0.1, "PER": 0.9},
                    {"O": 0.1, "PER": 0.9},
                ],
            ),
        ],
    )
    def test_merge_slices_for_single_document(
        self,
        list_slices: List[List[Union[str, Dict[str, float]]]],
        list_documents: List[Union[str, Dict[str, float]]],
    ):
        test_list_documents = merge_slices_for_single_document(list_slices)
        assert (
            test_list_documents == list_documents
        ), f"ERROR! test_list_documents = {test_list_documents} != {list_documents} = list_documents"

    @pytest.mark.parametrize(
        "tokens, predictions, token_predictions",
        [
            (
                [
                    "[CLS]",
                    "arbetsförmedl",
                    "##ingen",
                    "finns",
                    "i",
                    "stockholm",
                    "[SEP]",
                    "[PAD]",
                ],
                [
                    "[S]",
                    "ORG",
                    "ORG",
                    "O",
                    "O",
                    "O",
                    "[S]",
                    "[S]",
                ],
                [
                    ("arbetsförmedlingen", "ORG"),
                    ("finns", "O"),
                    ("i", "O"),
                    ("stockholm", "O"),
                ],
            ),
            (
                [
                    "[CLS]",
                    "arbetsförmedl",
                    "##ingen",
                    "finns",
                    "i",
                    "stockholm",
                    "[SEP]",
                    "[PAD]",
                ],
                [
                    "[S]",
                    "ORG",
                    "PER",
                    "O",
                    "O",
                    "O",
                    "[S]",
                    "[S]",
                ],
                [
                    ("arbetsförmedlingen", "ORG"),
                    ("finns", "O"),
                    ("i", "O"),
                    ("stockholm", "O"),
                ],
            ),
            (
                [
                    "[CLS]",
                    "1996",
                    "-",
                    "08",
                    "-",
                    "30",
                    "[PAD]",
                ],
                [
                    "[S]",
                    "ORG",
                    "ORG",
                    "O",
                    "ORG",
                    "O",
                    "[S]",
                ],
                [
                    ("1996", "ORG"),
                    ("-", "ORG"),
                    ("08", "O"),
                    ("-", "ORG"),
                    ("30", "O"),
                ],
            ),
        ],
    )
    def test_merge_subtoken_to_token_predictions(
        self,
        tokens: List[str],
        predictions: List[Union[str, Dict[str, float]]],
        token_predictions: List[Tuple[Union[str, Dict[str, float]]]],
    ):
        test_token_predictions = merge_subtoken_to_token_predictions(
            tokens, predictions
        )
        assert test_token_predictions == token_predictions, (
            f"test_token_predictions = "
            f"{test_token_predictions} != "
            f"{token_predictions}"
        )

    ####################################################################################################################
    @pytest.mark.parametrize(
        "word_predictions, input_text, word_predictions_restored",
        [
            # EXAMPLE 1 #######################
            (
                [
                    ("arbetsförmedlingen", "ORG"),
                    ("[UNK]", "O"),
                    ("i", "O"),
                    ("stockholm", "O"),
                ],
                "arbetsförmedlingen finns i stockholm",
                [
                    {
                        "char_start": "0",
                        "char_end": "18",
                        "token": "arbetsförmedlingen",
                        "tag": "ORG",
                    },
                    {
                        "char_start": "19",
                        "char_end": "24",
                        "token": "finns",
                        "tag": "O",
                    },
                    {"char_start": "25", "char_end": "26", "token": "i", "tag": "O"},
                    {
                        "char_start": "27",
                        "char_end": "36",
                        "token": "stockholm",
                        "tag": "O",
                    },
                ],
            ),
            # EXAMPLE 2 #######################
            (
                [
                    ("arbetsförmedlingen", "ORG"),
                    ("finns", "O"),
                    ("i", "O"),
                    ("[UNK]", "O"),
                ],
                "arbetsförmedlingen finns i stockholm",
                [
                    {
                        "char_start": "0",
                        "char_end": "18",
                        "token": "arbetsförmedlingen",
                        "tag": "ORG",
                    },
                    {
                        "char_start": "19",
                        "char_end": "24",
                        "token": "finns",
                        "tag": "O",
                    },
                    {"char_start": "25", "char_end": "26", "token": "i", "tag": "O"},
                    {
                        "char_start": "27",
                        "char_end": "36",
                        "token": "stockholm",
                        "tag": "O",
                    },
                ],
            ),
            # EXAMPLE 3 #######################
            (
                [
                    ("arbetsförmedlingen", "ORG"),
                    ("[UNK]", "O"),
                    ("i", "O"),
                    ("[UNK]", "O"),
                ],
                "arbetsförmedlingen finns i stockholm",
                [
                    {
                        "char_start": "0",
                        "char_end": "18",
                        "token": "arbetsförmedlingen",
                        "tag": "ORG",
                    },
                    {
                        "char_start": "19",
                        "char_end": "24",
                        "token": "finns",
                        "tag": "O",
                    },
                    {"char_start": "25", "char_end": "26", "token": "i", "tag": "O"},
                    {
                        "char_start": "27",
                        "char_end": "36",
                        "token": "stockholm",
                        "tag": "O",
                    },
                ],
            ),
            # EXAMPLE 4 #######################
            (
                [
                    ("Du", "O"),
                    ("behöver", "O"),
                    ("[UNK]", "O"),
                    ("kunna", "O"),
                    ("programmera", "B-SKILL"),
                ],
                "Du behöver ✓ kunna programmera",
                [
                    {
                        "char_start": "0",
                        "char_end": "2",
                        "token": "Du",
                        "tag": "O",
                    },
                    {
                        "char_start": "3",
                        "char_end": "10",
                        "token": "behöver",
                        "tag": "O",
                    },
                    {
                        "char_start": "11",
                        "char_end": "12",
                        "token": "✓",
                        "tag": "O",
                    },
                    {
                        "char_start": "13",
                        "char_end": "18",
                        "token": "kunna",
                        "tag": "O",
                    },
                    {
                        "char_start": "19",
                        "char_end": "30",
                        "token": "programmera",
                        "tag": "B-SKILL",
                    },
                ],
            ),
            # EXAMPLE 5 #######################
            (
                [
                    ("arbetsförmedlingen", "ORG"),
                    ("[UNK]", "O"),
                    ("i", "O"),
                    ("[UNK]", "O"),
                ],
                "arbetsförmedlingen íi i í",
                [
                    {
                        "char_start": "0",
                        "char_end": "18",
                        "token": "arbetsförmedlingen",
                        "tag": "ORG",
                    },
                    {
                        "char_start": "19",
                        "char_end": "21",
                        "token": "íi",
                        "tag": "O",
                    },
                    {"char_start": "22", "char_end": "23", "token": "i", "tag": "O"},
                    {
                        "char_start": "24",
                        "char_end": "25",
                        "token": "í",
                        "tag": "O",
                    },
                ],
            ),
            # EXAMPLE 6 #######################
            (
                [
                    ("arbetsförmedlingen", "ORG"),
                    ("[UNK]", "O"),
                    (".", "O"),
                    ("i", "O"),
                    ("[UNK]", "O"),
                ],
                "arbetsförmedlingen íi. i í",
                [
                    {
                        "char_start": "0",
                        "char_end": "18",
                        "token": "arbetsförmedlingen",
                        "tag": "ORG",
                    },
                    {
                        "char_start": "19",
                        "char_end": "21",
                        "token": "íi",
                        "tag": "O",
                    },
                    {"char_start": "21", "char_end": "22", "token": ".", "tag": "O"},
                    {"char_start": "23", "char_end": "24", "token": "i", "tag": "O"},
                    {
                        "char_start": "25",
                        "char_end": "26",
                        "token": "í",
                        "tag": "O",
                    },
                ],
            ),
            # EXAMPLE 7 #######################
            (
                [
                    ("arbetsförmedlingen", "ORG"),
                    ("[UNK]", "O"),
                    ("!", "O"),
                    ("i", "O"),
                    ("[UNK]", "O"),
                ],
                "arbetsförmedlingen íi! i í",
                [
                    {
                        "char_start": "0",
                        "char_end": "18",
                        "token": "arbetsförmedlingen",
                        "tag": "ORG",
                    },
                    {
                        "char_start": "19",
                        "char_end": "21",
                        "token": "íi",
                        "tag": "O",
                    },
                    {"char_start": "21", "char_end": "22", "token": "!", "tag": "O"},
                    {"char_start": "23", "char_end": "24", "token": "i", "tag": "O"},
                    {
                        "char_start": "25",
                        "char_end": "26",
                        "token": "í",
                        "tag": "O",
                    },
                ],
            ),
            # EXAMPLE 8 #######################
            (
                [
                    ("medarbetare", "O"),
                    ("i", "O"),
                    ("Stockholm", "O"),
                    (".", "O"),
                    ("[UNK]", "O"),
                    ("[NEWLINE]", "O"),
                    ("[NEWLINE]", "O"),
                    ("På", "O"),
                    ("Företaget", "O"),
                ],
                "medarbetare i Stockholm. ‍‍👨👩[NEWLINE][NEWLINE]På Företaget",
                [
                    {
                        "char_start": "0",
                        "char_end": "11",
                        "token": "medarbetare",
                        "tag": "O",
                    },
                    {
                        "char_start": "12",
                        "char_end": "13",
                        "token": "i",
                        "tag": "O",
                    },
                    {
                        "char_start": "14",
                        "char_end": "23",
                        "token": "Stockholm",
                        "tag": "O",
                    },
                    {
                        "char_start": "23",
                        "char_end": "24",
                        "token": ".",
                        "tag": "O",
                    },
                    {
                        "char_start": "25",
                        "char_end": "29",
                        "token": "‍‍👨👩",
                        "tag": "O",
                    },
                    {
                        "char_start": "29",
                        "char_end": "38",
                        "token": "[NEWLINE]",
                        "tag": "O",
                    },
                    {
                        "char_start": "38",
                        "char_end": "47",
                        "token": "[NEWLINE]",
                        "tag": "O",
                    },
                    {
                        "char_start": "47",
                        "char_end": "49",
                        "token": "På",
                        "tag": "O",
                    },
                    {
                        "char_start": "50",
                        "char_end": "59",
                        "token": "Företaget",
                        "tag": "O",
                    },
                ],
            ),
            #
        ],
    )
    def test_restore_unknown_tokens(
        self,
        word_predictions: List[
            Tuple[str, Union[str, Dict[str, float]]]
        ],  # List[Tuple[Union[str, Any], ...]],  # #
        input_text: str,
        word_predictions_restored: List[Dict[str, Any]],
    ):
        test_word_predictions_restored = restore_unknown_tokens(
            word_predictions,
            input_text,
            verbose=True,
        )
        assert test_word_predictions_restored == word_predictions_restored, (
            f"test_word_predictions_restored = "
            f"{test_word_predictions_restored} != "
            f"{word_predictions_restored}"
        )

    @pytest.mark.parametrize(
        "word_predictions, word_predictions_str",
        [
            (
                [
                    {
                        "char_start": 0,
                        "char_end": 18,
                        "token": "arbetsförmedlingen",
                        "tag": "ORG",
                    },
                    {
                        "char_start": 19,
                        "char_end": 24,
                        "token": "finns",
                        "tag": "O",
                    },
                ],
                [
                    {
                        "char_start": "0",
                        "char_end": "18",
                        "token": "arbetsförmedlingen",
                        "tag": "ORG",
                    },
                    {
                        "char_start": "19",
                        "char_end": "24",
                        "token": "finns",
                        "tag": "O",
                    },
                ],
            )
        ],
    )
    def test_assert_typing(
        self,
        word_predictions: List[Dict[str, Any]],
        word_predictions_str: List[Dict[str, str]],
    ):
        test_word_predictions_str = assert_typing(
            word_predictions,
        )
        assert test_word_predictions_str == word_predictions_str, (
            f"test_word_predictions_str = "
            f"{test_word_predictions_str} != "
            f"{word_predictions_str}"
        )



from typing import Optional, Dict, List, Tuple, Any
from os.path import join, isfile
import pandas as pd
from mlflow.entities import Run
from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.utils.util_functions import epoch2checkpoint
from nerblackbox.modules.utils.util_functions import (
    get_run_name,
    compute_mean_and_dmean,
)


class ExperimentResults:

    """class that contains results of a single experiment."""

    METRICS = {
        "EPOCH_BEST": "EPOCH_BEST",
        "EPOCH_STOPPED": "EPOCH_STOPPED",
        "EPOCH_BEST_VAL_ENTITY_FIL_F1_MICRO": "VAL_ENT_F1",
        "EPOCH_BEST_TEST_ENTITY_FIL_F1_MICRO": "TEST_ENT_F1",
        "EPOCH_BEST_VAL_TOKEN_FIL_F1_MICRO": "VAL_TOK_F1",
        "EPOCH_BEST_TEST_TOKEN_FIL_F1_MICRO": "TEST_TOK_F1",
        "entity_fil_precision_micro": "TEST_ENT_PRE",
        "entity_fil_recall_micro": "TEST_ENT_REC",
        "entity_fil_asr_abidance": "TEST_ENT_ASR_ABI",
        "entity_fil_asr_precision_micro": "TEST_ENT_ASR_PRE",
        "entity_fil_asr_recall_micro": "TEST_ENT_ASR_REC",
        "entity_fil_asr_f1_micro": "TEST_ENT_ASR_F1",
    }
    METRICS_PLUS = dict(**{"EPOCHS": "EPOCHS"}, **METRICS)
    PARAMS = {
        "max_seq_length": "max_seq",
        "lr_schedule": "lr_sch",
        "batch_size": "batch_sz",
    }

    def __init__(
        self,
        _id: Optional[str] = None,
        name: Optional[str] = None,
        experiment: Optional[pd.DataFrame] = None,
        single_runs: Optional[pd.DataFrame] = None,
        average_runs: Optional[pd.DataFrame] = None,
        best_single_run: Optional[Dict] = None,
        best_average_run: Optional[Dict] = None,
    ):
        """
        Args:
            _id: e.g. '1'
            name: e.g. 'my_experiment'
            experiment: overview on experiment parameters
            single_runs: overview on run parameters & single  results
            average_runs: overview on run parameters & average results
            best_single_run: overview on best run parameters & single results
            best_average_run: overview on best run parameters & average results
        """
        self._id: str = _id if _id is not None else ""
        self.name: str = name if name is not None else ""
        self.experiment: pd.DataFrame = (
            experiment if experiment is not None else pd.DataFrame()
        )
        self.single_runs: pd.DataFrame = (
            single_runs if single_runs is not None else pd.DataFrame()
        )
        self.average_runs: pd.DataFrame = (
            average_runs if average_runs is not None else pd.DataFrame()
        )
        self.best_single_run: Dict = (
            best_single_run if best_single_run is not None else dict()
        )
        self.best_average_run: Dict = (
            best_average_run if best_average_run is not None else dict()
        )

    @classmethod
    def from_mlflow_runs(
        cls, _runs: List[Run], _experiment_id: str, _experiment_name: str
    ) -> "ExperimentResults":
        """
        Args:
            _runs: [List of mlflow.entities.Run]
            _experiment_id: [str], e.g. '0'
            _experiment_name: [str], e.g. 'my_experiment'

        Returns:
            ExperimentResults instance
        """
        experiment_results = ExperimentResults(
            _id=_experiment_id, name=_experiment_name
        )
        experiment_results._parse_and_create_dataframe(
            _runs
        )  # attr: experiment, single_runs, average_runs
        experiment_results.extract_best_single_run()  # attr: best_single_run
        experiment_results.extract_best_average_run()  # attr: best_average_run

        return experiment_results

    ####################################################################################################################
    # 1. PARSE AND CREATE DATAFRAME -> experiment, single_runs, average_runs
    ####################################################################################################################
    def _parse_and_create_dataframe(
        self,
        _runs: List[Run],
    ) -> None:
        r"""
        turn mlflow Run objects (= search_runs() results) into dataframes

        Args:
            _runs:   [list] of [mlflow.entities.Run objects]

        Created Attr:
            experiment:   [pandas DataFrame] overview on experiment parameters
            single_runs:  [pandas DataFrame] overview on single  run parameters & results
            average_runs: [pandas DataFrame] overview on average run parameters & results
        """
        ###########################################
        # parameters_experiment & parameters_runs
        ###########################################
        parameters_runs, parameters_experiment = self._parse_runs(_runs)

        ###########################################
        # rename
        ###########################################
        parameters_runs_renamed = self._rename_parameters_runs(parameters_runs)

        ###########################################
        # average
        ###########################################
        parameters_runs_renamed_average = self._average(parameters_runs_renamed)

        ###########################################
        # dataframes
        ###########################################
        self.experiment = pd.DataFrame(parameters_experiment, index=["experiment"]).T
        self.single_runs, self.average_runs = self._2_sorted_dataframes(
            parameters_runs_renamed, parameters_runs_renamed_average
        )

    @classmethod
    def _parse_runs(
        cls, _runs: List[Run]
    ) -> Tuple[Dict[Tuple[str, str], Any], Dict[str, Any]]:
        """
        Args:
            _runs: list of mlflow.entities.Run objects

        Returns:
            _parameters_runs:       information on single runs
            _parameters_experiment: information on whole experiment
        """
        _parameters_runs: Dict[Tuple[str, str], Any] = dict()
        _parameters_experiment: Dict[str, Any] = dict()
        for i in range(len(_runs)):
            if len(_runs[i].data.metrics) == 0:  # experiment
                _parameters_experiment = {
                    k: [v] for k, v in _runs[i].data.params.items()
                }
            else:  # run
                if ("info", "run_id") not in _parameters_runs.keys():
                    _parameters_runs[("info", "run_id")] = [_runs[i].info.run_id]
                else:
                    _parameters_runs[("info", "run_id")].append(_runs[i].info.run_id)

                if ("info", "run_name_nr") not in _parameters_runs.keys():
                    _parameters_runs[("info", "run_name_nr")] = [
                        _runs[i].data.tags["mlflow.runName"]
                    ]
                else:
                    _parameters_runs[("info", "run_name_nr")].append(
                        _runs[i].data.tags["mlflow.runName"]
                    )

                for k, v in _runs[i].data.params.items():
                    if ("params", k) not in _parameters_runs.keys():
                        _parameters_runs[("params", k)] = [v]
                    else:
                        _parameters_runs[("params", k)].append(v)

                for k in cls.METRICS_PLUS.keys():
                    if ("metrics", k) not in _parameters_runs.keys():
                        try:
                            _parameters_runs[("metrics", k)] = [
                                _runs[i].data.metrics[k]
                            ]
                        except:
                            _parameters_runs[("metrics", k)] = [-1]
                    else:
                        try:
                            _parameters_runs[("metrics", k)].append(
                                _runs[i].data.metrics[k]
                            )
                        except:
                            _parameters_runs[("metrics", k)] = [-1]

        for k in ["epoch_best".upper(), "epoch_stopped".upper()]:
            try:
                _parameters_runs[("metrics", k)] = [
                    int(elem) for elem in _parameters_runs[("metrics", k)]
                ]
            except:
                _parameters_runs[("metrics", k)] = [-1]

        if _parameters_runs[("metrics", "epoch_best".upper())] != [-1]:
            _parameters_runs[("metrics", "epochs".upper())] = [
                elem + 1 for elem in _parameters_runs[("metrics", "epoch_best".upper())]
            ]

        return _parameters_runs, _parameters_experiment

    @classmethod
    def _rename_parameters_runs_single(cls, _tuple: Tuple[str, str]) -> Tuple[str, str]:
        """
        Args:
            _tuple: e.g. ("metrics", "D_entity_fil_precision_micro")

        Returns:
            _tuple_renamed: ("metrics", "TEST_ENT_PRE")
        """
        _category, _field = _tuple
        if _category == "params":
            return (
                "params",
                cls.PARAMS[_field] if _field in cls.PARAMS.keys() else _field,
            )
        elif _category == "metrics":
            if _field in cls.METRICS_PLUS.keys():
                return "metrics", cls.METRICS_PLUS[_field]
            elif _field.replace("D_", "") in cls.METRICS_PLUS.keys():
                return "metrics", f"D_{cls.METRICS_PLUS[_field.replace('D_', '')]}"
            else:
                raise Exception(f"ERROR! _rename_parameters_runs_single() failed.")
        else:
            return _category, _field

    @classmethod
    def _rename_parameters_runs(
        cls, _parameters_runs: Dict[Tuple[str, str], Any]
    ) -> Dict[Tuple[str, str], Any]:
        return {
            cls._rename_parameters_runs_single(k): v
            for k, v in _parameters_runs.items()
        }

    @classmethod
    def _average(
        cls, _parameters_runs_renamed: Dict[Tuple[str, str], Any]
    ) -> Dict[Tuple[str, str], List[str]]:
        """
        Args:
            _parameters_runs_renamed:         information on single runs

        Returns:
            _parameters_runs_renamed_average: information on averaged single runs
        """
        keys = (
            [("info", "run_name")]
            + [
                k
                for k in _parameters_runs_renamed.keys()
                if k[0] not in ["info", "metrics"]
            ]
            + [
                ("metrics", k)
                for k in ["CONVERGENCE"] + list(cls.METRICS_PLUS.values())
            ]
        )
        _parameters_runs_renamed_average: Dict[Tuple[str, str], List[str]] = {
            k: list() for k in keys
        }

        runs_name_nr = _parameters_runs_renamed[("info", "run_name_nr")]
        nr_runs = len(runs_name_nr)
        runs_name = list(
            set([get_run_name(run_name_nr) for run_name_nr in runs_name_nr])
        )
        run_name_2_indices = {
            run_name: [
                idx
                for idx in range(nr_runs)
                if get_run_name(runs_name_nr[idx]) == run_name
            ]
            for run_name in runs_name
        }

        #######################
        # loop over runs_name
        #######################
        for run_name in runs_name:
            # add ('info', 'run_name')
            _parameters_runs_renamed_average[("info", "run_name")].append(run_name)

            indices = run_name_2_indices[run_name]

            # add ('params', *)
            random_index = indices[0]
            keys_kept = [k for k in _parameters_runs_renamed.keys() if k[0] == "params"]
            for key in keys_kept:
                _parameters_runs_renamed_average[key].append(
                    _parameters_runs_renamed[key][random_index]
                )

            # add ('metrics', *)
            metrics = dict()

            indices_converged = [
                idx
                for idx in indices
                if _parameters_runs_renamed[
                    ("metrics", cls.METRICS["EPOCH_BEST_VAL_ENTITY_FIL_F1_MICRO"])
                ][idx]
                > 0
            ]
            convergence_str = f"{len(indices_converged)}/{len(indices)}"
            _parameters_runs_renamed_average[("metrics", "CONVERGENCE")].append(
                convergence_str
            )

            for _metric in cls.METRICS_PLUS.values():
                _mean, _dmean = cls._get_mean_and_dmean(
                    indices_converged,
                    _parameters_runs_renamed,
                    _metric=_metric,
                )
                metrics[_metric] = f"{_mean:.5f} +- "
                metrics[_metric] += f"{_dmean:.5f}" if _dmean is not None else "###"
            for k in metrics.keys():
                key = ("metrics", k)
                _parameters_runs_renamed_average[key].append(metrics[k])

        return _parameters_runs_renamed_average

    @classmethod
    def _get_mean_and_dmean(
        cls,
        _indices: List[int],
        _parameters_runs_renamed: Dict[Tuple[str, str], Any],
        _metric: str,
    ) -> Tuple[float, Optional[float]]:
        try:
            values = [
                _parameters_runs_renamed[("metrics", _metric)][idx] for idx in _indices
            ]
        except IndexError:
            values = [-1 for _ in _indices]
        return compute_mean_and_dmean(values)

    def _2_sorted_dataframes(
        self,
        _parameters_runs_renamed: Dict[Tuple[str, str], Any],
        _parameters_runs_renamed_average: Dict[Tuple[str, str], Any],
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        by = ("metrics", self.METRICS["EPOCH_BEST_VAL_ENTITY_FIL_F1_MICRO"])

        # single runs
        try:
            _single_runs = pd.DataFrame(_parameters_runs_renamed).sort_values(
                by=by, ascending=False
            )
        except:
            _single_runs = None

        # average runs
        try:
            _average_runs = pd.DataFrame(_parameters_runs_renamed_average).sort_values(
                by=by, ascending=False
            )
        except:
            _average_runs = None

        return _single_runs, _average_runs

    ####################################################################################################################
    # 2a. EXTRACT BEST SINGLE RUN -> best_single_run
    ####################################################################################################################
    def extract_best_single_run(self) -> None:
        if self.experiment is not None and self.single_runs is not None:
            _df_best_single_run = self.single_runs.iloc[0, :]

            assert (
                self.name is not None
            ), f"ERROR! self.name is None, extract_best_single_run() failed."
            checkpoint = join(
                env_variable("DIR_CHECKPOINTS"),
                self.name,
                _df_best_single_run[("info", "run_name_nr")],
                epoch2checkpoint(_df_best_single_run[("metrics", "EPOCH_BEST")]),
            )

            fields_info = ["run_id", "run_name_nr"]

            self.best_single_run = dict(
                **{
                    "exp_id": self._id,
                    "exp_name": self.name,
                    "checkpoint": checkpoint if isfile(checkpoint) else None,
                },
                **{
                    field: _df_best_single_run[("info", field)] for field in fields_info
                },
                **{
                    field: _df_best_single_run[("metrics", field)]
                    for field in self.METRICS_PLUS.values()
                },
            )
        else:
            self.best_single_run = dict()

    ####################################################################################################################
    # 2b. EXTRACT BEST AVERAGE RUN -> best_average_run
    ####################################################################################################################
    def extract_best_average_run(self) -> None:
        if self.experiment is not None and self.average_runs is not None:
            _df_best_average_run = self.average_runs.iloc[0, :]

            self.best_average_run = dict(
                **{
                    "exp_id": self._id,
                    "exp_name": self.name,
                    "run_name": _df_best_average_run[("info", "run_name")],
                },
                **{
                    field: _df_best_average_run[("metrics", field)]
                    for field in list(self.METRICS_PLUS.values()) + ["CONVERGENCE"]
                },
            )
        else:
            self.best_average_run = dict()



from itertools import product
from typing import Optional, Dict, Union, List, Any
from nerblackbox.modules.utils.parameters import PARAMS, HPARAMS
from nerblackbox.modules.experiment_config.experiment_config import ExperimentConfig
from nerblackbox.modules.utils.util_functions import get_run_name_nr


class Experiment:
    def __init__(
        self,
        experiment_name: str,
        from_config: bool = True,
        run_name: Optional[str] = None,
        device: str = "gpu",
        fp16: bool = True,
    ):
        """
        Args:
            experiment_name: e.g. 'exp1'
            from_config:     whether to read experiment config from file
            run_name:        e.g. 'runA'
            device:          e.g. 'gpu'
            fp16:
        """
        self.experiment_name = experiment_name
        self.from_config = from_config
        self.run_name = run_name
        self.device = device
        self.fp16 = fp16

        self.exp_default = ExperimentConfig("default")
        self.exp = ExperimentConfig(self.experiment_name)

        # info for runs
        self.runs_name_nr: List[str] = list()
        self.runs_params: Dict[str, Dict[str, Union[str, int, float, bool]]] = dict()
        self.runs_hparams: Dict[str, Dict[str, Union[str, int, float, bool]]] = dict()
        self._create_info_for_runs()

        # info for mlflow
        self.params_and_hparams: Dict[str, Dict[str, Any]]
        self._create_info_for_mlflow()

    def _create_info_for_runs(self) -> None:
        """
        parse <experiment_name>.ini files
        if self.run_name is specified, parse only that run. else parse all runs.

        Created Attr:
            runs_name_nr [list] of [str],
                         e.g. ['runA-1', 'runA-2', 'runB-1', 'runB-2']
            runs_params  [dict] w/ keys = run [str], values = params [dict],
                         e.g. {'runA-1': {'patience': 2, 'mode': 'min', ..}, ..}
            runs_hparams [dict] w/ keys = run [str], values = hparams [dict],
                         e.g. {'runA-1': {'lr_max': 2e-5, 'max_epochs': 20, ..}, ..}
        """
        if (
            "params" in self.exp.config.keys()
            and "multiple_runs" in self.exp.config["params"].keys()
        ):
            multiple_runs = self.exp.config["params"]["multiple_runs"]
        elif (
            "params" in self.exp_default.config.keys()
            and "multiple_runs" in self.exp_default.config["params"].keys()
        ):
            multiple_runs = self.exp_default.config["params"]["multiple_runs"]
        else:
            raise Exception(
                f"multiple runs is neither specified in the experiment config nor in the default config."
            )

        # _params_config & _hparams_config
        _runs_params: Dict[str, Union[str, int, float, bool]] = dict()
        _runs_hparams: Dict[str, Union[str, int, float, bool]] = dict()

        if self.run_name is None:  # multiple runs
            run_names = self.exp.run_names
        else:
            run_names = [
                run_name for run_name in self.exp.run_names if run_name == self.run_name
            ]
            assert (
                len(run_names) == 1
            ), f"ERROR! found {len(run_names)} runs with name = {self.run_name}"

        for run_name, run_nr in product(
            run_names, list(range(1, int(multiple_runs) + 1))
        ):
            run_name_nr = get_run_name_nr(run_name, run_nr)

            # _run_params
            _run_params: Dict[str, Union[str, int, float, bool]] = {
                "experiment_name": self.experiment_name,
                "from_config": self.from_config,
                "run_name": run_name,
                "run_name_nr": run_name_nr,
                "device": self.device,
                "fp16": self.fp16,
                "experiment_run_name_nr": f"{self.experiment_name}/{run_name_nr}",
            }
            _run_params.update(self.exp_default.config["params"])
            _run_params.update(self.exp.config["params"])

            # _run_hparams
            _run_hparams: Dict[str, Union[str, int, float, bool]] = dict()
            _run_hparams.update(self.exp_default.config["hparams"])
            _run_hparams.update(self.exp.config["hparams"])

            for k, v in self.exp.config[run_name].items():
                if k in PARAMS:
                    _run_params[k] = v
                elif k in HPARAMS:
                    _run_hparams[k] = v
                else:
                    raise Exception(f"parameter = {k} is unknown.")

            self.runs_name_nr.append(run_name_nr)
            self.runs_params[run_name_nr] = _run_params
            self.runs_hparams[run_name_nr] = _run_hparams

        assert set(self.runs_params.keys()) == set(self.runs_hparams.keys())

    def _create_info_for_mlflow(self) -> None:
        """
        get dictionary of all parameters & their values that belong to
        either generally    to experiment (key = "general")
        or     specifically to run        (key = run_name)

        Create Attr:
            params_and_hparams: [dict] e.g. {'patience': 2, 'mode': 'min', ..}

        Returns: -
        """

        self.params_and_hparams = {
            "general": {},
        }
        self.params_and_hparams["general"].update(self.exp_default.config["params"])
        self.params_and_hparams["general"].update(self.exp_default.config["hparams"])
        self.params_and_hparams["general"].update(self.exp.config["params"])
        self.params_and_hparams["general"].update(self.exp.config["hparams"])

        for run_name in self.exp.run_names:
            self.params_and_hparams[run_name] = self.exp.config[run_name]



from os.path import join
import pytorch_lightning as pl
from pytorch_lightning.core.optimizer import LightningOptimizer
from abc import ABC, abstractmethod
import torch
from torch.optim.optimizer import Optimizer
from typing import List, Dict, Optional, Callable, Tuple, Any, Union
from omegaconf import DictConfig

from torch.optim.lr_scheduler import LambdaLR

from transformers import AdamW
from transformers import get_linear_schedule_with_warmup
from transformers import get_constant_schedule_with_warmup
from transformers import get_cosine_schedule_with_warmup
from transformers import get_cosine_with_hard_restarts_schedule_with_warmup
from transformers import AutoTokenizer
from transformers import PreTrainedModel

from nerblackbox.modules.ner_training.data_preprocessing.data_preprocessor import (
    DataPreprocessor,
)
from nerblackbox.modules.utils.util_functions import split_parameters
from nerblackbox.modules.ner_training.ner_model_evaluation import NerModelEvaluation
from nerblackbox.modules.ner_training.annotation_tags.annotation import Annotation
from nerblackbox.modules.ner_training.logging.mlflow_client import MLflowClient
from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger
from nerblackbox.modules.utils.env_variable import env_variable


class NerModel(pl.LightningModule, ABC):
    def __init__(self, hparams: DictConfig):
        """
        :param hparams: attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
        """
        super().__init__()
        self.save_hyperparameters(hparams)

        # split up hparams
        (
            self.params,
            self.hyperparameters,
            self.log_dirs,
            self.experiment,
        ) = split_parameters(hparams)

        # preparations
        self._preparations()

    ####################################################################################################################
    # Abstract Base Methods ############################################################################################
    ####################################################################################################################
    @abstractmethod
    def _preparations(self):
        """
        :created attr: default_logger    [DefaultLogger]
        :created attr: logged_metrics    [LoggedMetrics]
        :created attr: tokenizer         [transformers AutoTokenizer]
        :created attr: encoding          [dict] e.g. {"\n": "[NEWLINE]", "\t": "[TAB]"}
        :created attr: special_tokens    [list] of [str] e.g. ["[NEWLINE]", "[TAB]"]
        :created attr: data_preprocessor [DataPreprocessor]
        :created attr: annotation        [Annotation]
        :created attr: model             [transformers AutoModelForTokenClassification]

        [only train]
        :created attr: mlflow_client          [MLflowClient]
        :created attr: epoch_metrics          [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: classification_reports [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: pretrained_model_name  [str]
        :created attr: dataloader             [dict] w/ keys 'train', 'val', 'test' & values = [torch Dataloader]
        :created attr: optimizer              [torch optimizer]
        :created attr: scheduler              [torch LambdaLR]
        :return: -
        """
        self.default_logger: DefaultLogger
        self.annotation: Annotation
        self.model: PreTrainedModel
        self.mlflow_client: MLflowClient
        self.optimizer: Optimizer
        self.scheduler: LambdaLR

    def _preparations_data_general(self):
        """
        :created attr: pretrained_model_name  [str]
        :created attr: tokenizer              [transformers AutoTokenizer]
        :created attr: data_preprocessor      [DataPreprocessor]
        :return: -
        """
        # 1. pretrained model name
        try:
            # use transformers model
            AutoTokenizer.from_pretrained(
                self.params.pretrained_model_name, use_fast=False
            )
            self.pretrained_model_name = self.params.pretrained_model_name
        except ValueError:
            # use local model
            self.pretrained_model_name = join(
                env_variable("DATA_DIR"),
                "pretrained_models",
                self.params.pretrained_model_name,
            )

        # 2. tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.pretrained_model_name,
            do_lower_case=False,
            additional_special_tokens=self.special_tokens,
            use_fast=True,
        )  # do_lower_case needs to be False !!

        # 3. data_preprocessor
        self.data_preprocessor = DataPreprocessor(
            tokenizer=self.tokenizer,
            do_lower_case=self.params.uncased,  # can be True !!
            max_seq_length=self.hyperparameters.max_seq_length,
            default_logger=self.default_logger,
        )

    ####################################################################################################################
    # TRAIN/VAL/TEST
    ####################################################################################################################
    ####################################################################################################################
    # FORWARD & BACKWARD PROPAGATION
    ####################################################################################################################
    def forward(self, *args, **kwargs) -> List:
        """
        Args:
            args = (batch),
            where batch = Dict[str, torch.Tensor]
            with keys = subset of EncodingsKeys,
                 values = 2D torch tensor of shape [batch_size, seq_length]
            e.g.
            input_ids      = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..], [..], ..]
            attention_mask = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
            token_type_ids = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
            labels         = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..], [..], ..]

        Returns:
            _outputs: [list] of 1 or 2 elements:
                   i)  _loss: [float] cross entropy between _labels & _tags_ids_predictions
                                      on non-padding tokens (i.e. where elements in _input_ids are not 0)
                   ii) _labels_prediction_logits: [torch tensor] of shape [batch_size, seq_length, vocabulary_size],
                       if labels are provided in _batch
        """
        return self.model(**args[0])

    ####################################################################################################################
    # TRAIN
    ####################################################################################################################
    def training_step(self, batch, batch_idx):
        # REQUIRED
        input_ids, attention_mask, token_type_ids, labels = self._parse_batch(batch)
        outputs = self(batch)
        batch_train_loss = outputs[0]

        # logging
        self._write_metrics_for_tensorboard("train", {"all_loss": batch_train_loss})

        # debug
        if batch_idx == 0:
            self._debug_step_check(
                "train",
                batch,
                outputs,
                input_ids,
                attention_mask,
                token_type_ids,
                labels,
            )

        return {"loss": batch_train_loss}

    ####################################################################################################################
    # OPTIMIZER
    ####################################################################################################################
    def configure_optimizers(self):
        # REQUIRED
        # can return multiple optimizers and learning_rate schedulers
        return self.optimizer

    def optimizer_step(
        self,
        epoch: int,
        batch_idx: int,
        optimizer: Union[Optimizer, LightningOptimizer],
        optimizer_idx: int = 0,
        optimizer_closure: Optional[Callable] = None,
        on_tpu: bool = False,
        using_native_amp: bool = False,
        using_lbfgs: bool = False,
    ) -> None:
        assert isinstance(
            optimizer, Optimizer
        ), f"ERROR! type(optimizer) = {type(optimizer)} should be Optimizer."

        # update params
        if optimizer is not None:
            optimizer.step(closure=optimizer_closure)
            optimizer.zero_grad()

        # update learning rate
        self.scheduler.step()

    ####################################################################################################################
    # VALID
    ####################################################################################################################
    def validation_step(self, batch, batch_idx):
        # OPTIONAL
        input_ids, attention_mask, token_type_ids, labels = self._parse_batch(batch)
        outputs = self(batch)
        batch_loss, logits = outputs[:2]

        # debug
        if batch_idx == 0:
            self._debug_step_check(
                "val", batch, outputs, input_ids, attention_mask, token_type_ids, labels
            )

        return batch_loss, labels, logits

    def validation_epoch_end(self, outputs):
        """
        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_labels, batch_logits] for each batch
        :return:        [dict] w/ key 'val_loss' & value = mean batch loss of val dataset [float]
        """
        # OPTIONAL
        assert self.trainer is not None, f"ERROR! self.trainer is None."

        if self.trainer.state.fn == "fit":
            metrics = "loss"
        elif self.trainer.state.fn == "validate":
            metrics = "all"
        else:
            raise Exception(
                f"ERROR! self.trainer.state.fn = {self.trainer.state.fn} unexpected."
            )

        return self._validate_on_epoch("val", outputs=outputs, metrics=metrics)

    ####################################################################################################################
    # TEST
    ####################################################################################################################
    def test_step(self, batch, batch_idx):
        # OPTIONAL
        input_ids, attention_mask, token_type_ids, labels = self._parse_batch(batch)
        outputs = self(batch)
        batch_loss, logits = outputs[:2]

        # debug
        if batch_idx == 0:
            self._debug_step_check(
                "test",
                batch,
                outputs,
                input_ids,
                attention_mask,
                token_type_ids,
                labels,
            )

        return batch_loss, labels, logits

    def test_epoch_end(self, outputs):
        """
        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_labels, batch_logits] for each batch
        :return:        [dict] w/ key 'test_loss' & value = mean batch loss of test dataset [float]
        """
        # OPTIONAL
        return self._validate_on_epoch("test", outputs=outputs, metrics="all")

    ####################################################################################################################
    # DATALOADER
    ####################################################################################################################
    def train_dataloader(self):
        # REQUIRED
        return self.dataloader["train"]

    def val_dataloader(self):
        # OPTIONAL
        return self.dataloader["val"]

    def test_dataloader(self):
        # OPTIONAL
        return self.dataloader["test"]

    ####################################################################################################################
    # HELPER METHODS
    # 1. PREPARATIONS
    # 2. VALIDATE / COMPUTE METRICS
    # 3. PRINT / LOG
    ####################################################################################################################

    ####################################################################################################################
    # 1. PREPARATIONS
    ####################################################################################################################
    def _create_optimizer(
        self,
        learning_rate: float,
        fp16: bool = True,
        no_decay: Tuple[str, ...] = ("bias", "LayerNorm.weight"),
    ) -> Optimizer:
        """
        create optimizer with basic learning rate and L2 normalization for some parameters

        Args:
            learning_rate: [float] basic learning rate
            fp16:          [bool]
            no_decay:      [tuple of str] parameters that contain one of those are not subject to L2 normalization

        Returns:
            optimizer:   [torch optimizer]
        """
        # Remove unused pooler that otherwise break Apex
        param_optimizer = list(self.model.named_parameters())
        optimizer_grouped_parameters: List[Any] = [
            {
                "params": [
                    p for n, p in param_optimizer if not any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.01,
            },
            {
                "params": [
                    p for n, p in param_optimizer if any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.0,
            },
        ]
        # self.default_logger.log_debug(
        #     '> param_optimizer:',
        #     [n for n, p in param_optimizer],
        # )
        self.default_logger.log_debug(
            "> parameters w/  weight decay:",
            len(optimizer_grouped_parameters[0]["params"])
            if isinstance(optimizer_grouped_parameters[0]["params"], list)
            else None,
        )
        self.default_logger.log_debug(
            "> parameters w/o weight decay:",
            len(optimizer_grouped_parameters[1]["params"])
            if isinstance(optimizer_grouped_parameters[1]["params"], list)
            else None,
        )
        if fp16:
            optimizer = AdamW(optimizer_grouped_parameters, lr=learning_rate)
            # optimizer = FusedAdam(optimizer_grouped_parameters, lr=self.learning_rate, bias_correction=False)
            # optimizer = FP16_Optimizer(optimizer, dynamic_loss_scale=True)

        else:
            optimizer = AdamW(optimizer_grouped_parameters, lr=learning_rate)
            # optimizer = FusedAdam(optimizer_grouped_parameters, lr=learning_rate)

        # optimizer = BertAdam(optimizer_grouped_parameters,lr=2e-5, warmup=.1)
        return optimizer

    def _create_scheduler(
        self, _lr_warmup_epochs, _lr_schedule, _lr_max_epochs=None, _lr_num_cycles=None
    ) -> LambdaLR:
        """
        create scheduler with warmup
        ----------------------------
        :param _lr_warmup_epochs:   [int]
        :param _lr_schedule:        [str], 'linear', 'constant', 'cosine', 'cosine_with_hard_resets', 'hybrid'
        :param _lr_max_epochs:      [int, optional] needed for _lr_schedule != 'constant'
        :param _lr_num_cycles:      [float, optional], e.g. 0.5, 1.0, only for cosine learning rate schedules
        :return: scheduler          [torch LambdaLR] learning rate scheduler
        """
        if _lr_schedule not in [
            "constant",
            "linear",
            "cosine",
            "cosine_with_hard_restarts",
            "hybrid",
        ]:
            raise Exception(f"lr_schedule = {_lr_schedule} not implemented.")

        num_warmup_steps = self._get_steps(_lr_warmup_epochs)

        scheduler_params = {
            "num_warmup_steps": num_warmup_steps,
            "last_epoch": -1,
        }

        if _lr_schedule in ["constant", "hybrid"]:
            return get_constant_schedule_with_warmup(self.optimizer, **scheduler_params)
        else:
            assert (
                _lr_max_epochs is not None
            ), f"ERROR! need to specify _lr_max_epochs for _lr_schedule = {_lr_schedule}"
            num_training_steps = self._get_steps(_lr_max_epochs)
            scheduler_params["num_training_steps"] = num_training_steps

            if _lr_schedule == "linear":
                return get_linear_schedule_with_warmup(
                    self.optimizer, **scheduler_params
                )
            else:
                if _lr_num_cycles is not None:
                    scheduler_params[
                        "num_cycles"
                    ] = _lr_num_cycles  # else: use default values

                if _lr_schedule == "cosine":
                    scheduler_params["num_training_steps"] = num_training_steps
                    return get_cosine_schedule_with_warmup(
                        self.optimizer, **scheduler_params
                    )
                elif _lr_schedule == "cosine_with_hard_restarts":
                    scheduler_params["num_training_steps"] = num_training_steps
                    return get_cosine_with_hard_restarts_schedule_with_warmup(
                        self.optimizer, **scheduler_params
                    )
                else:
                    raise Exception(
                        "create scheduler: logic is broken."
                    )  # this should never happen

    def _get_steps(self, _num_epochs):
        """
        helper method for _create_scheduler
        gets steps = num_epochs * (number of training data samples)
        -----------------------------------------------------------------
        :param _num_epochs: [int], e.g. 10
        :return: steps:     [int], e.g. 2500 (in case of 250 training data samples)
        """
        return _num_epochs * len(self.dataloader["train"])

    @staticmethod
    def _parse_batch(_batch) -> Tuple[torch.Tensor, ...]:
        """
        Args:
            _batch: Dict with keys = subset of EncodingsKeys, values = 2D torch tensor of shape [batch_size, seq_length]
            e.g.
            input_ids      = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..], [..], ..]
            attention_mask = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
            token_type_ids = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
            labels         = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..], [..], ..]

        Returns:
            input_ids      = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..], [..], ..]
            attention_mask = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
            token_type_ids = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
            labels         = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..], [..], ..]
        """
        input_ids = _batch["input_ids"]
        attention_mask = _batch["attention_mask"]
        token_type_ids = (
            _batch["token_type_ids"] if "token_type_ids" in _batch.keys() else None
        )
        labels = _batch["labels"]
        return input_ids, attention_mask, token_type_ids, labels

    ####################################################################################################################
    # 2. VALIDATE / COMPUTE METRICS
    ####################################################################################################################
    def _validate_on_epoch(
        self,
        phase: str,
        outputs: List[torch.Tensor],
        metrics: str,
    ) -> Dict[str, float]:
        """
        Args:
            phase:   [str], 'val', 'test'
            outputs: [list] of [lists] = [batch_loss, batch_labels, batch_logits] with 3 torch tensors for each batch
            metrics: [str], 'loss', 'all'

        Returns:
            [dict] w/ key '<phase>_loss' & value = mean batch loss [float]
        """
        if metrics == "loss":
            # CPU (numpy)
            # batch_loss = [output[0].detach().cpu().numpy() for output in outputs]
            # epoch_loss = np.stack(batch_loss).mean()

            # GPU (pytorch)
            batch_loss = [output[0].detach() for output in outputs]
            epoch_loss = torch.mean(torch.stack(batch_loss)).item()  # -> float

            self.log(f"{phase}_loss", epoch_loss)  # for early stopping callback
        elif metrics == "all":
            ner_model_evaluation = NerModelEvaluation(
                current_epoch=self.current_epoch,
                annotation=self.annotation,
                default_logger=self.default_logger,
                logged_metrics=self.logged_metrics,
            )
            (
                epoch_metrics,
                classification_report,
                confusion_matrix,
                epoch_loss,
            ) = ner_model_evaluation.execute(phase, outputs)
            self._log_metrics_confusion_matrix_classification_report(
                phase,
                epoch_metrics,
                confusion_matrix,
                classification_report,
            )
        else:
            raise Exception(f"ERROR! metrics = {metrics} unexpected.")

        return {f"{phase}_loss": epoch_loss}

    def _log_metrics_confusion_matrix_classification_report(
        self,
        phase: str,
        epoch_metrics: Dict[str, float],
        confusion_matrix: Optional[str] = "",
        classification_report: Optional[str] = "",
    ) -> None:
        """
        Args:
            phase:                  'val', 'test'
            epoch_metrics           keys 'all_acc', 'fil_f1_micro', .. & values = [flaot]
            confusion_matrix:
            classification_report:
        """
        # tracked metrics & classification reports
        self._add_epoch_metrics(
            phase, self.current_epoch, epoch_metrics
        )  # attr: epoch_metrics

        # logging: tb
        self._write_metrics_for_tensorboard(phase, epoch_metrics)

        # logging: mlflow
        if phase == "test":
            self.mlflow_client.log_metrics(self.current_epoch, epoch_metrics)

            if confusion_matrix is not None:
                self.mlflow_client.log_artifact(
                    confusion_matrix,
                    overwrite=True,
                )
            if classification_report is not None:
                self.mlflow_client.log_artifact(
                    classification_report,
                    overwrite=False,
                )

            self.mlflow_client.finish_artifact_mlflow()

        # print
        self._print_metrics(phase, epoch_metrics, classification_report)

        self.default_logger.log_debug(f"--> {phase}: epoch done")

    def _add_epoch_metrics(self, phase, epoch, _epoch_metrics):
        """
        add _epoch_metrics to attribute/dict epoch_<phase>_metrics
        --------------------------------------------------------------
        :param: epoch:                [int]
        :param: _epoch_metrics:       [dict] w/ keys 'loss', 'acc', 'f1' & values = [np array]
        :changed attr: epoch_metrics: [dict] w/ keys = 'val'/'test and
                                                values = dict w/ keys = epoch [int], values = _epoch_metrics [dict]
        :return: -
        """
        self.epoch_metrics[phase][epoch] = _epoch_metrics

    def _write_metrics_for_tensorboard(self, phase, metrics):
        """
        write metrics for tensorboard
        -----------------------------
        :param phase:         [str] 'train' or 'val'
        :param metrics:       [dict] w/ keys 'loss', 'acc', 'f1_macro_all', 'f1_micro_all'
        :return: -
        """
        # tb_logs: all
        tb_logs = {
            f'{phase}/{k.split("_", 1)[0]}/{k.split("_", 1)[1]}': v
            for k, v in metrics.items()
        }

        # tb_logs: learning rate
        if phase == "train":
            tb_logs[f"{phase}/learning_rate"] = self.scheduler.get_last_lr()[0]

        # tb_logs
        self.logger.log_metrics(tb_logs, self.global_step)

    ####################################################################################################################
    # 3. PRINT / LOG
    ####################################################################################################################
    def _debug_step_check(
        self,
        phase,
        _batch,
        _outputs,
        _input_ids,
        _attention_mask,
        _token_type_ids,
        _labels,
    ):
        self.default_logger.log_debug(f"{phase.upper()} STEP CHECK")
        self.default_logger.log_debug(f"batch on gpu:   {_batch['input_ids'].is_cuda}")
        self.default_logger.log_debug(f"outputs on gpu: {_outputs[0].is_cuda}")
        name_tensor_combinations = [
            ("input_ids", _input_ids),
            ("attention_mask", _attention_mask),
            ("token_type_ids", _token_type_ids),
            ("labels", _labels),
        ]
        for _name, _tensor in name_tensor_combinations:
            if _tensor is not None:
                self.default_logger.log_debug(
                    f"{_name.ljust(15)} shape|1st row: {_tensor.shape} | {_tensor[0]}"
                )
            else:
                self.default_logger.log_debug(f"{_name.ljust(15)} = None")

    def _print_metrics(self, phase, _metrics, _classification_reports: Optional[str]):
        """
        :param phase:         [str] 'train' or 'val'
        :param _metrics:
        :param _classification_reports:
        :return:
        """
        self.default_logger.log_info("")
        self.default_logger.log_info(
            f"--- Epoch #{self.current_epoch} {phase.ljust(5).upper()} ----"
        )
        self.default_logger.log_info(
            "token  all loss:         {:.2f}".format(_metrics["token_all_loss"])
        )
        self.default_logger.log_info(
            "token  all acc:          {:.2f}".format(_metrics["token_all_acc"])
        )
        self.default_logger.log_info(
            "token  all f1 (micro):   {:.2f}".format(_metrics["token_all_f1_micro"])
        )
        self.default_logger.log_info(
            "token  fil f1 (micro):   {:.2f}".format(_metrics["token_fil_f1_micro"])
        )
        self.default_logger.log_info(
            "entity fil f1 (micro):   {:.2f}".format(_metrics["entity_fil_f1_micro"])
        )
        self.default_logger.log_info(f"-----------------------")
        if _classification_reports is not None:
            self.default_logger.log_debug(_classification_reports)



import os
from os.path import join
from configparser import ConfigParser
from nerblackbox.modules.utils.parameters import PARAMS, HPARAMS
from nerblackbox.modules.utils.env_variable import env_variable
from typing import Union, Tuple, Any, Dict, List


class ExperimentConfig:
    """
    class that parses <experiment_name>.ini files
    """

    def __init__(self, experiment_name: str):
        """
        Args:
            experiment_name: e.g. 'exp1', 'default
        """
        self.config: Dict[str, Dict[str, str]]
        self.run_names: List[str]
        self.config, self.run_names = self._get_config(experiment_name)

    ####################################################################################################################
    # HELPER METHODS
    ####################################################################################################################
    def _get_config(
        self, experiment_name: str
    ) -> Tuple[Dict[str, Dict[str, str]], List[str]]:
        """
        get ConfigParser instance and derive config dictionary from it

        Args:
            experiment_name: e.g. 'exp1', 'default

        Returns:
            _config_dict: w/ keys = sections [str], values = [dict] w/ key: value = params: values
            _run_names: e.g. ["runA", "runB"]
        """
        config_path = join(
            env_variable("DIR_EXPERIMENT_CONFIGS"), f"{experiment_name}.ini"
        )
        if not os.path.isfile(config_path):
            raise Exception(f"config file at {config_path} does not exist")

        _config = ConfigParser()
        _config.read(config_path)
        _config_dict: Dict[str, Dict[str, Any]] = {
            s: dict(_config.items(s)) for s in _config.sections()
        }  # {'hparams': {'monitor': 'val_loss'}}
        _config_dict = {
            s: {k: self._convert(k, v) for k, v in subdict.items()}
            for s, subdict in _config_dict.items()
        }

        # combine sections 'dataset', 'model' & 'settings' to single section 'params'
        _config_dict["params"] = dict()
        for s in ["dataset", "model", "settings"]:
            if s in _config_dict.keys():
                _config_dict["params"].update(_config_dict[s])
                _config_dict.pop(s)

        # derive uncased
        if (
            "uncased" not in _config_dict["params"].keys()
            and "pretrained_model_name" in _config_dict["params"]
        ):
            if "uncased" in _config_dict["params"]["pretrained_model_name"]:
                _config_dict["params"]["uncased"] = True
            elif "cased" in _config_dict["params"]["pretrained_model_name"]:
                _config_dict["params"]["uncased"] = False
            else:
                _config_dict["params"]["uncased"] = False
                print(
                    "ATTENTION! could not derive uncased = True/False from pretrained_model_name."
                    " => assume model is cased"
                )

        _run_names = [
            run_name for run_name in _config.sections() if run_name.startswith("run")
        ]

        return _config_dict, _run_names

    @staticmethod
    def _convert(_input_key: str, _input_value: str) -> Union[str, int, float, bool]:
        """
        convert _input string to str/int/float/bool

        Args:
            _input_key:        [str],              e.g. 'lr_schedule' or 'prune_ratio_train' or 'checkpoints'
            _input_value:      [str],              e.g. 'constant'    or '0.01'              or 'False'

        Returns:
            converted_input: [str/int/float/bool], e.g. 'constant'    or  0.01               or  False
        """
        if _input_key in PARAMS.keys():
            convert_to = PARAMS[_input_key]
        elif _input_key in HPARAMS.keys():
            convert_to = HPARAMS[_input_key]
        else:
            raise Exception(f"_input_key = {_input_key} unknown.")

        if convert_to == "str":
            return _input_value
        elif convert_to == "int":
            return int(_input_value)
        elif convert_to == "float":
            return float(_input_value)
        elif convert_to == "bool":
            return _input_value not in ["False", "false"]
        else:  # pragma: no cover
            raise Exception(f"convert_to = {convert_to} not known.")



from pytorch_lightning.callbacks.base import Callback


class LearningRateChanger(Callback):
    def __init__(self, max_epochs: int, lr_cooldown_epochs: int):
        self.max_epochs = max_epochs
        self.lr_cooldown_epochs = lr_cooldown_epochs
        self.change_epoch = (
            self.max_epochs - self.lr_cooldown_epochs - 1
        )  # minus 1 as epoch counting starts from 0
        super().__init__()

    def on_validation_end(self, trainer, pl_module):
        if trainer.current_epoch == self.change_epoch:
            print(
                f"> LearningRateChanger: change LR to linear cool-down with {self.lr_cooldown_epochs} epochs."
            )
            pl_module.scheduler = pl_module._create_scheduler(
                _lr_warmup_epochs=0,
                _lr_schedule="linear",
                _lr_max_epochs=self.lr_cooldown_epochs,
            )



from copy import deepcopy
from typing import List, Tuple, Dict, Union, Optional

EncodeDecodeMappings = List[Tuple[int, str, str]]
Predictions = List[Dict[str, Union[str, Dict]]]


class TextEncoder:
    def __init__(
        self,
        encoding: Dict[str, str],
        model_special_tokens: Optional[List[str]] = None,
    ):
        r"""
        Examples:
            ```
            TextEncoder(
                encoding={"\n": "[NEWLINE]", "\t": "[TAB]"},
                model_special_tokens=["[NEWLINE]", "[TAB]"],
            )
            ```
        Args:
            encoding: mapping to special tokens
            model_special_tokens: special tokens that the model was trained on
        """
        self.encoding = encoding
        if model_special_tokens is None:
            print(
                "ATTENTION! DID NOT CHECK THAT MODEL WAS TRAINED WITH SPECIAL TOKENS ACCORDING TO encoding! "
                "It is recommended to provide an 'model_special_tokens' argument to check."
            )
        else:
            assert sorted(list(set(encoding.values()))) == sorted(
                list(set(model_special_tokens))
            ), (
                f"ERROR! encoding values = {sorted(list(set(encoding.values())))} does not equal "
                f"model_special_tokens = {sorted(list(set(model_special_tokens)))}"
            )

    def encode(
        self, text_list: List[str]
    ) -> Tuple[List[str], List[EncodeDecodeMappings]]:
        r"""
        encodes list of text using self.encoding

        Examples:
            ```
            text_encoded_list, encode_decode_mappings_list = encode(text_list=["an\n example"])
            # text_encoded_list           = ["an[NEWLINE] example"]
            # encode_decode_mappings_list = [[(2, "\n", "[NEWLINE]")]]
            ```

        Args:
            text_list: original text

        Returns:
            text_encoded_list: encoded text

            encode_decode_mappings_list: mappings (char_start, original token, encoded token)
        """
        list_of_single_encodings = [self._encode_single(text) for text in text_list]
        return [elem[0] for elem in list_of_single_encodings], [
            elem[1] for elem in list_of_single_encodings
        ]

    def decode(
        self,
        text_encoded_list: List[str],
        encode_decode_mappings_list: List[EncodeDecodeMappings],
        predictions_encoded_list: List[Predictions],
    ) -> Tuple[List[str], List[Predictions]]:
        r"""
        decodes list of text_encoded and predictions_encoded using encode_decode_mappings

        Examples:
            ```
            text_list, predictions_list = decode(
                text_encoded_list=["an[NEWLINE] example"],
                encode_decode_mappings_list=[[(2, "\n", "[NEWLINE]")]]),
                predictions_encoded_list=[[{"char_start": "12", "char_end": "19", "token": "example", "tag": "TAG"}]]
            )
            # text_list = ["an\n example"]
            # predictions_list = [[{"char_start": "4", "char_end": "11", "token": "example", "tag": "TAG"}]]
            ```

        Args:
            text_encoded_list: encoded text

            encode_decode_mappings_list: mappings (char_start, original token, encoded token)

            predictions_encoded_list: encoded predictions

        Returns:
            text_list: original / decoded text

            predictions_list: original / decoded predictions
        """
        list_of_single_decodings = [
            self._decode_single(
                text_encoded, encode_decode_mappings, predictions_encoded
            )
            for text_encoded, encode_decode_mappings, predictions_encoded in zip(
                text_encoded_list, encode_decode_mappings_list, predictions_encoded_list
            )
        ]
        return [elem[0] for elem in list_of_single_decodings], [
            elem[1] for elem in list_of_single_decodings
        ]

    def _encode_single(self, text: str) -> Tuple[str, EncodeDecodeMappings]:
        r"""
        encodes single text using self.encoding

        Args:
            text: e.g. "an\n example"

        Returns:
            text_encoded: e.g. "an[NEWLINE] example"
            encode_decode_mappings: e.g. [(2, "\n", "[NEWLINE]")]
        """
        text_encoded = deepcopy(text)
        encode_decode_mappings = list()
        for k, v in self.encoding.items():
            while k in text_encoded:
                index = text_encoded.find(k)
                text_encoded = text_encoded.replace(k, v, 1)
                encode_decode_mappings.append((index, k, v))
        encode_decode_mappings.reverse()
        return text_encoded, encode_decode_mappings

    @staticmethod
    def _decode_single(
        text_encoded: str,
        encode_decode_mappings: EncodeDecodeMappings,
        predictions_encoded: Predictions,
    ) -> Tuple[str, Predictions]:
        r"""
        decodes single text_encoded and predictions_encoded using encode_decode_mapping

        Args:
            text_encoded: e.g. "an[NEWLINE] example"
            encode_decode_mappings: e.g. [(2, "\n", "[NEWLINE]")]
            predictions_encoded: e.g. [{"char_start": "12", "char_end": "19", "token": "example", "tag": "TAG"}]

        Returns:
            text: e.g. "an\n example"
            predictions: e.g. [{"char_start": "4", "char_end": "11", "token": "example", "tag": "TAG"}]
        """
        text = deepcopy(text_encoded)
        predictions = deepcopy(predictions_encoded)
        for encode_decode_mapping in encode_decode_mappings:
            index, k, v = encode_decode_mapping
            # print(index, k, v)
            assert (
                text[index : index + len(v)] == v
            ), f"ERROR! text[{index}:{index + len(v)}] = {text[index:index + len(v)]} != {v}"
            text = text[:index] + k + text[index + len(v) :]

            for prediction in predictions:
                assert isinstance(prediction["char_start"], str) and isinstance(
                    prediction["char_end"], str
                ), f"ERROR! expected str, got type ({prediction['char_start']}, {prediction['char_end']})"
                if int(prediction["char_start"]) == index and int(
                    prediction["char_end"]
                ) == index + len(v):
                    prediction["char_end"] = str(
                        int(prediction["char_end"]) - len(v) + len(k)
                    )
                    prediction["token"] = k
                elif int(prediction["char_end"]) > index:
                    prediction["char_start"] = str(
                        int(prediction["char_start"]) - len(v) + len(k)
                    )
                    prediction["char_end"] = str(
                        int(prediction["char_end"]) - len(v) + len(k)
                    )

        return text, predictions



import pytest
import numpy as np
import torch
from typing import Dict, List, Union, Optional
from nerblackbox.modules.ner_training.ner_model_evaluation import NerModelEvaluation
from nerblackbox.modules.ner_training.metrics.logged_metrics import LoggedMetrics
from nerblackbox.tests.utils import PseudoDefaultLogger
from nerblackbox.tests.test_utils import pytest_approx
from nerblackbox.modules.ner_training.annotation_tags.annotation import Annotation


class TestNerModelEvaluation:

    ner_model_evaluation = NerModelEvaluation(
        current_epoch=2,
        annotation=Annotation(["O", "ORG", "PER"]),
        default_logger=PseudoDefaultLogger(),
        logged_metrics=LoggedMetrics(),
    )
    ner_model_evaluation_bio = NerModelEvaluation(
        current_epoch=2,
        annotation=Annotation(["O", "B-ORG", "B-PER", "I-ORG", "I-PER"]),
        default_logger=PseudoDefaultLogger(),
        logged_metrics=LoggedMetrics(),
    )

    ####################################################################################################################
    # TEST #############################################################################################################
    ####################################################################################################################
    @pytest.mark.parametrize(
        "outputs, " "np_batch, " "np_epoch",
        [
            (
                [
                    [
                        torch.tensor([1.0, 2.0]),
                        torch.tensor([3, 4]),
                        torch.tensor([[9.0, 1.2, 3.3], [0.7, 0.6, 5.6]]),
                    ],
                    [
                        torch.tensor([2.0, 3.0]),
                        torch.tensor([5, 6]),
                        torch.tensor([[0.9, 2.1, 3.3], [7.0, 6.0, 6.5]]),
                    ],
                ],
                {
                    "loss": [
                        np.array([1.0, 2.0], dtype=np.float32),
                        np.array([2.0, 3.0], dtype=np.float32),
                    ],
                    "tag_ids": [
                        np.array([3, 4], dtype=np.float32),
                        np.array([5, 6], dtype=np.float32),
                    ],
                    "logits": [
                        np.array([[9.0, 1.2, 3.3], [0.7, 0.6, 5.6]], dtype=np.float32),
                        np.array([[0.9, 2.1, 3.3], [7.0, 6.0, 6.5]], dtype=np.float32),
                    ],
                },
                {
                    "loss": 2.0,
                    "tag_ids": np.array([3, 4, 5, 6], dtype=np.float32),
                    "logits": np.array(
                        [
                            [9.0, 1.2, 3.3],
                            [0.7, 0.6, 5.6],
                            [0.9, 2.1, 3.3],
                            [7.0, 6.0, 6.5],
                        ],
                        dtype=np.float32,
                    ),
                },
            ),
        ],
    )
    def test_convert_and_combine(
        self,
        outputs: List[torch.Tensor],
        np_batch: Dict[str, List[np.ndarray]],
        np_epoch: Dict[str, Union[np.number, np.ndarray]],
    ) -> None:
        # 1. output -> np_batch
        test_np_batch = self.ner_model_evaluation._convert_output_to_np_batch(outputs)
        for key in np_batch.keys():
            for _test, _true in zip(test_np_batch[key], np_batch[key]):
                assert np.array_equal(
                    _test, _true
                ), f"test_np_batch[{key}] = {_test} != {_true}"

        # 2. np_batch -> np_epoch
        test_np_epoch = self.ner_model_evaluation._combine_np_batch_to_np_epoch(
            np_batch
        )
        for key in np_epoch.keys():
            if key == "loss":
                assert (
                    test_np_epoch[key] == np_epoch[key]
                ), f"test_np_epoch[{key}] = {test_np_epoch[key]} != {np_epoch[key]}"
            else:
                assert np.array_equal(
                    test_np_epoch[key], np_epoch[key]
                ), f"test_np_epoch[{key}] = {test_np_epoch[key]} != {np_epoch[key]}"

    @pytest.mark.parametrize(
        "_np_tag_ids, " "_np_logits, " "true_flat, " "pred_flat",
        [
            (
                np.array([[0, 1], [1, 0]]),
                np.array([[[0.8, 0.2], [0.8, 0.2]], [[0.3, 0.7], [0.3, 0.7]]]),
                np.array([0, 1, 1, 0]),
                np.array([0, 0, 1, 1]),
            ),
            (
                np.array([0, 1, 1, 1]),
                np.array(
                    [
                        [9.0, 1.2, 3.3],
                        [0.7, 0.6, 5.6],
                        [0.9, 2.1, 3.3],
                        [7.0, 6.0, 6.5],
                    ]
                ),
                np.array([0, 1, 1, 1], dtype=int),
                np.array([0, 2, 2, 0], dtype=int),
            ),
        ],
    )
    def test_reduce_and_flatten(
        self,
        _np_tag_ids: np.ndarray,
        _np_logits: np.ndarray,
        true_flat: np.ndarray,
        pred_flat: np.ndarray,
    ) -> None:
        test_true_flat, test_pred_flat = self.ner_model_evaluation._reduce_and_flatten(
            _np_tag_ids, _np_logits
        )
        assert np.array_equal(
            test_true_flat, true_flat
        ), f"true_flat: test = {test_true_flat} != {true_flat} = true"
        assert np.array_equal(
            test_pred_flat, pred_flat
        ), f"pred_flat: test = {test_true_flat} != {pred_flat} = true"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "_np_tag_ids, " "tags",
        [
            (
                [0, 2, 1, 2],
                ["O", "PER", "ORG", "PER"],
            ),
            (
                [-100, 2, 1, -100],
                ["[S]", "PER", "ORG", "[S]"],
            ),
            (
                [0, -100, 1, 2],
                ["O", "[S]", "ORG", "PER"],
            ),
        ],
    )
    def test_convert_tag_ids_to_tags(
        self,
        _np_tag_ids: List[int],
        tags: List[str],
    ) -> None:
        test_tags = self.ner_model_evaluation._convert_tag_ids_to_tags(_np_tag_ids)
        assert np.array_equal(test_tags, tags), f"test = {test_tags} != {tags} = true"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "_tags, " "tags_new",
        [
            (
                {
                    "true": ["[S]", "PER", "ORG", "[S]"],
                    "pred": ["PER", "ORG", "O", "O"],
                },
                {
                    "true": ["PER", "ORG"],
                    "pred": ["ORG", "O"],
                },
            ),
            (
                {
                    "true": ["PER", "[S]", "ORG", "[S]"],
                    "pred": ["PER", "ORG", "O", "O"],
                },
                {
                    "true": ["PER", "ORG"],
                    "pred": ["PER", "O"],
                },
            ),
        ],
    )
    def test_get_rid_of_special_tag_occurrences(
        self,
        _tags: Dict[str, List[str]],
        tags_new: Dict[str, List[str]],
    ) -> None:
        test_tags_new = self.ner_model_evaluation._get_rid_of_special_tag_occurrences(
            _tags
        )
        for k in tags_new.keys():
            assert np.array_equal(
                test_tags_new[k], tags_new[k]
            ), f"key = {k}: test = {test_tags_new[k]} != {tags_new[k]} = true"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "_tag_subset, _tags_plain, _filtered_classes, _filtered_class_index",
        [
            (
                "all",
                {"true": ["O", "ORG", "PER"], "pred": []},
                ["O", "ORG", "PER"],
                None,
            ),
            (
                "fil",
                {"true": ["O", "ORG", "PER"], "pred": []},
                ["ORG", "PER"],
                None,
            ),
            (
                "O",
                {"true": ["O", "ORG", "PER"], "pred": []},
                ["O"],
                None,
            ),
            (
                "PER",
                {
                    "true": ["O", "ORG", "PER"],
                    "pred": ["O", "PER", "ORG"],
                },
                ["PER"],
                1,
            ),
            (
                "PER",
                {
                    "true": ["O", "PER", "ORG"],
                    "pred": ["O", "ORG", "PER"],
                },
                ["PER"],
                1,
            ),
            (
                "ORG",
                {
                    "true": ["O", "ORG", "PER"],
                    "pred": ["O", "PER", "ORG"],
                },
                ["ORG"],
                0,
            ),
            (
                "ORG",
                {
                    "true": ["O", "PER", "ORG"],
                    "pred": ["O", "ORG", "PER"],
                },
                ["ORG"],
                0,
            ),
            (
                "all",
                {"true": ["ORG"], "pred": []},
                ["O", "ORG", "PER"],
                None,
            ),
            (
                "fil",
                {"true": ["ORG"], "pred": []},
                ["ORG", "PER"],
                None,
            ),
            (
                "O",
                {"true": ["ORG"], "pred": []},
                ["O"],
                None,
            ),
            (
                "PER",
                {"true": ["ORG"], "pred": []},
                ["PER"],
                None,
            ),
            (
                "ORG",
                {"true": ["ORG"], "pred": []},
                ["ORG"],
                0,
            ),
            (
                "PER",
                {"true": ["PER"], "pred": []},
                ["PER"],
                0,
            ),
            (
                "ORG",
                {"true": ["PER"], "pred": []},
                ["ORG"],
                None,
            ),
        ],
    )
    def test_get_filtered_classes(
        self,
        _tag_subset: str,
        _tags_plain: Dict[str, List[str]],
        _filtered_classes: List[str],
        _filtered_class_index: Optional[int],
    ) -> None:
        (
            test_filtered_classes,
            test_filtered_class_index,
        ) = self.ner_model_evaluation._get_filtered_classes(_tag_subset, _tags_plain)

        assert (
            test_filtered_classes == _filtered_classes
        ), f"test_filtered_classes = {test_filtered_classes} != {_filtered_classes}"
        assert (
            test_filtered_class_index == _filtered_class_index
        ), f"test_filtered_class_index = {test_filtered_class_index} != {_filtered_class_index}"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "annotation_scheme, " "_tags, " "_phase, " "tag_subset, " "metrics",
        [
            (
                "plain",
                {
                    "true": ["O", "PER", "ORG", "PER"],
                    "pred": ["O", "PER", "ORG", "PER"],
                },
                "test",
                "fil",
                {
                    "entity_fil_numberofclasses_macro": 2,
                    "entity_fil_recall_micro": 1.0,
                    "entity_fil_recall_macro": 1.0,
                    "entity_fil_precision_micro": 1.0,
                    "entity_fil_precision_macro": 1.0,
                    "entity_fil_f1_micro": 1.0,
                    "entity_fil_f1_macro": 1.0,
                    "entity_fil_asr_abidance": 1.0,
                    "entity_fil_asr_precision_micro": 1.0,
                    "entity_fil_asr_recall_micro": 1.0,
                    "entity_fil_asr_f1_micro": 1.0,
                    "token_fil_numberofclasses_macro": 2,
                    "token_fil_recall_micro": 1.0,
                    "token_fil_recall_macro": 1.0,
                    "token_fil_precision_micro": 1.0,
                    "token_fil_precision_macro": 1.0,
                    "token_fil_f1_micro": 1.0,
                    "token_fil_f1_macro": 1.0,
                },
            ),
            (
                "plain",
                {
                    "true": ["O", "PER", "ORG", "PER"],
                    "pred": ["O", "PER", "ORG", "PER"],
                },
                "test",
                "PER",
                {
                    "entity_PER_recall": 1.0,
                    "entity_PER_precision": 1.0,
                    "entity_PER_f1": 1.0,
                    "token_PER_recall": 1.0,
                    "token_PER_precision": 1.0,
                    "token_PER_f1": 1.0,
                },
            ),
            (
                "bio",
                {
                    "true": ["O", "B-PER", "I-PER", "O", "B-PER"],
                    "pred": ["O", "B-PER", "I-PER", "I-PER", "B-PER"],
                },
                "test",
                "PER",
                {
                    "entity_PER_recall": 0.5,
                    "entity_PER_precision": 0.5,
                    "entity_PER_f1": 0.5,
                    "token_PER_recall": 1.0,
                    "token_PER_precision": 0.75,
                    "token_PER_f1": 0.86,
                },
            ),
            (
                "bio",
                {
                    "true": ["O", "B-PER", "B-ORG", "O"],
                    "pred": ["O", "B-ORG", "I-ORG", "O"],
                },
                "test",
                "PER",
                {
                    "entity_PER_recall": 0.0,
                    "entity_PER_precision": -1.0,
                    "entity_PER_f1": -1.0,
                    "token_PER_recall": 0.0,
                    "token_PER_precision": -1.0,
                    "token_PER_f1": -1.0,
                },
            ),
            (
                "bio",
                {
                    "true": ["O", "B-PER", "B-ORG", "O"],
                    "pred": ["O", "B-ORG", "I-ORG", "O"],
                },
                "test",
                "ORG",
                {
                    "entity_ORG_recall": 0.0,
                    "entity_ORG_precision": 0.0,
                    "entity_ORG_f1": 0.0,
                    "token_ORG_recall": 1.0,
                    "token_ORG_precision": 0.5,
                    "token_ORG_f1": 0.66,
                },
            ),
            (
                "bio",
                {
                    "true": ["O", "B-PER", "B-ORG", "O"],
                    "pred": ["O", "B-ORG", "B-ORG", "O"],
                },
                "test",
                "PER",
                {
                    "entity_PER_recall": 0.0,
                    "entity_PER_precision": -1.0,
                    "entity_PER_f1": -1.0,
                    "token_PER_recall": 0.0,
                    "token_PER_precision": -1.0,
                    "token_PER_f1": -1.0,
                },
            ),
        ],
    )
    def test_compute_metrics_for_tags_subset(
        self,
        annotation_scheme: str,
        _tags: Dict[str, List[str]],
        _phase: str,
        tag_subset: str,
        metrics: Dict[str, float],
    ) -> None:
        if annotation_scheme == "plain":
            test_metrics = self.ner_model_evaluation._compute_metrics_for_tags_subset(
                _tags, _phase, tag_subset
            )
        elif annotation_scheme == "bio":
            test_metrics = (
                self.ner_model_evaluation_bio._compute_metrics_for_tags_subset(
                    _tags, _phase, tag_subset
                )
            )
        else:
            raise Exception(f"annotation_scheme = {annotation_scheme} unknown.")

        a = set(list(test_metrics.keys()))
        b = set(list(metrics.keys()))
        assert len(a.difference(b)) == 0, a.difference(b)
        assert len(b.difference(a)) == 0, b.difference(a)

        for k in list(set(list(test_metrics.keys()) + list(metrics.keys()))):
            assert test_metrics[k] == pytest_approx(
                metrics[k]
            ), f"test_metrics[{k}] = {test_metrics[k]} != {metrics[k]}"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "annotation_scheme, " "phase, " "np_epoch, " "epoch_metrics, " "epoch_tags",
        [
            (
                "plain",
                "test",
                {
                    "loss": 2.0,
                    "tag_ids": np.array(
                        [0, 2, 1, 0], dtype=np.float32
                    ),  # [O, PER, ORG, O]  (true)
                    "logits": np.array(
                        [  # O, ORG, PER
                            [9.0, 3.3, 1.2],  # O
                            [0.7, 5.6, 0.6],  # ORG
                            [0.9, 3.3, 2.1],  # ORG
                            [7.0, 6.5, 6.0],  # O
                        ],
                        dtype=np.float32,
                    ),
                },
                {
                    "entity_fil_numberofclasses_macro": 1,
                    "entity_fil_precision_micro": 0.0,
                    "entity_fil_precision_macro": 0.0,
                    "entity_fil_recall_micro": 0.0,
                    "entity_fil_recall_macro": 0.0,
                    "entity_fil_f1_micro": 0.0,
                    "entity_fil_f1_macro": 0.0,
                    "entity_fil_asr_abidance": 1.0,
                    "entity_fil_asr_precision_micro": 0.0,
                    "entity_fil_asr_recall_micro": 0.0,
                    "entity_fil_asr_f1_micro": 0.0,
                    "entity_PER_precision": -1.0,
                    "entity_PER_recall": 0.0,
                    "entity_PER_f1": -1.0,
                    "entity_ORG_precision": 0.0,
                    "entity_ORG_recall": 0.0,
                    "entity_ORG_f1": 0.0,
                    "token_all_loss": 2.0,
                    "token_all_acc": 0.75,
                    "token_all_precision_micro": 0.75,
                    "token_all_precision_macro": 0.50,
                    "token_all_recall_micro": 0.75,
                    "token_all_recall_macro": 0.66,
                    "token_all_f1_micro": 0.75,
                    "token_all_f1_macro": 0.55,
                    "token_fil_numberofclasses_macro": 1,
                    "token_fil_precision_micro": 0.5,
                    "token_fil_precision_macro": 0.25,
                    "token_fil_recall_micro": 0.5,
                    "token_fil_recall_macro": 0.5,
                    "token_fil_f1_micro": 0.5,
                    "token_fil_f1_macro": 0.33,
                    "token_O_precision": 1.0,
                    "token_O_recall": 1.0,
                    "token_O_f1": 1.0,
                    "token_PER_precision": -1.0,
                    "token_PER_recall": 0.0,
                    "token_PER_f1": -1.0,
                    "token_ORG_precision": 0.5,
                    "token_ORG_recall": 1.0,
                    "token_ORG_f1": 0.66,
                },
                {
                    "true": np.array(["O", "PER", "ORG", "O"]),
                    "pred": np.array(["O", "ORG", "ORG", "O"]),
                },
            ),
            (
                "bio",
                "test",
                {
                    "loss": 2.0,
                    "tag_ids": np.array(
                        [0, 2, 1, 0], dtype=np.float32
                    ),  # [O, B-PER, B-ORG, O]  (true)
                    "logits": np.array(
                        [
                            # 0, B-ORG, B-PER, I-ORG, I-PER  (pred)
                            [9.0, 3.3, 1.2, 1.0, 1.0],  # O
                            [0.7, 5.6, 1.2, 1.0, 1.0],  # B-ORG
                            [0.9, 1.0, 2.1, 3.3, 1.0],  # I-ORG
                            [7.0, 6.5, 6.9, 1.0, 1.0],  # O
                        ],
                        dtype=np.float32,
                    ),
                },
                {
                    "entity_fil_numberofclasses_macro": 1,
                    "entity_fil_precision_micro": 0.0,
                    "entity_fil_precision_macro": 0.0,
                    "entity_fil_recall_micro": 0.0,
                    "entity_fil_recall_macro": 0.0,
                    "entity_fil_f1_micro": 0.0,
                    "entity_fil_f1_macro": 0.0,
                    "entity_fil_asr_abidance": 1.0,
                    "entity_fil_asr_precision_micro": 0.0,
                    "entity_fil_asr_recall_micro": 0.0,
                    "entity_fil_asr_f1_micro": 0.0,
                    "entity_PER_precision": -1.0,
                    "entity_PER_recall": 0.0,
                    "entity_PER_f1": -1.0,
                    "entity_ORG_precision": 0.0,
                    "entity_ORG_recall": 0.0,
                    "entity_ORG_f1": 0.0,
                    "token_all_loss": 2.0,
                    "token_all_acc": 0.75,
                    "token_all_precision_micro": 0.75,
                    "token_all_precision_macro": 0.5,
                    "token_all_recall_micro": 0.75,
                    "token_all_recall_macro": 0.66,
                    "token_all_f1_micro": 0.75,
                    "token_all_f1_macro": 0.55,
                    "token_fil_numberofclasses_macro": 1,
                    "token_fil_precision_micro": 0.5,
                    "token_fil_precision_macro": 0.25,
                    "token_fil_recall_micro": 0.5,
                    "token_fil_recall_macro": 0.5,
                    "token_fil_f1_micro": 0.5,
                    "token_fil_f1_macro": 0.33,
                    "token_O_precision": 1.0,
                    "token_O_recall": 1.0,
                    "token_O_f1": 1.0,
                    "token_PER_precision": -1.0,
                    "token_PER_recall": 0.0,
                    "token_PER_f1": -1.0,
                    "token_ORG_precision": 0.5,
                    "token_ORG_recall": 1.0,
                    "token_ORG_f1": 0.66,
                },
                {
                    "true": np.array(["O", "B-PER", "B-ORG", "O"]),
                    "pred": np.array(["O", "B-ORG", "I-ORG", "O"]),
                },
            ),
        ],
    )
    def test_compute_metrics(
        self,
        annotation_scheme: str,
        phase: str,
        np_epoch: Dict[str, Union[float, np.ndarray]],
        epoch_metrics: Dict[str, float],
        epoch_tags: Dict[str, np.ndarray],
    ) -> None:
        if annotation_scheme == "plain":
            (
                test_epoch_metrics,
                test_epoch_tags,
            ) = self.ner_model_evaluation._compute_metrics(phase, np_epoch)
        elif annotation_scheme == "bio":
            (
                test_epoch_metrics,
                test_epoch_tags,
            ) = self.ner_model_evaluation_bio._compute_metrics(phase, np_epoch)
        else:
            raise Exception(f"annotation_scheme = {annotation_scheme} unknown.")

        a = set(list(test_epoch_metrics.keys()))
        b = set(list(epoch_metrics.keys()))
        assert len(a.difference(b)) == 0, a.difference(b)
        assert len(b.difference(a)) == 0, b.difference(a)

        for k in list(set(list(test_epoch_tags.keys()) + list(epoch_tags.keys()))):
            assert np.array_equal(
                test_epoch_tags[k], epoch_tags[k]
            ), f"key = {k}: test_epoch_tags = {test_epoch_tags[k]} != {epoch_tags[k]}"

        for k in list(
            set(list(test_epoch_metrics.keys()) + list(epoch_metrics.keys()))
        ):
            assert test_epoch_metrics[k] == pytest_approx(
                epoch_metrics[k]
            ), f"key = {k}: test_epoch_metrics = {test_epoch_metrics[k]} != {epoch_metrics[k]}"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "phase, " "outputs, " "epoch_metrics, " "epoch_loss",
        [
            (
                "test",
                [
                    [
                        torch.tensor([1.0, 2.0]),
                        torch.tensor([0, 2]),  # =          [O, PER]  (true)
                        torch.tensor(
                            [[9.0, 1.2, 3.3], [0.7, 5.6, 0.6]]
                        ),  # = [0, 1] = [O, ORG]  (pred)
                    ],
                    [
                        torch.tensor([2.0, 3.0]),
                        torch.tensor([1, 0]),  # =          [ORG, O]  (true)
                        torch.tensor(
                            [[0.9, 3.3, 2.1], [7.0, 6.5, 6.0]]
                        ),  # = [1, 0] = [ORG, O]  (pred)
                    ],
                ],
                {
                    "entity_fil_numberofclasses_macro": 1,
                    "entity_fil_precision_micro": 0.0,
                    "entity_fil_precision_macro": 0.0,
                    "entity_fil_recall_micro": 0.0,
                    "entity_fil_recall_macro": 0.0,
                    "entity_fil_f1_micro": 0.0,
                    "entity_fil_f1_macro": 0.0,
                    "entity_fil_asr_abidance": 1.0,
                    "entity_fil_asr_precision_micro": 0.0,
                    "entity_fil_asr_recall_micro": 0.0,
                    "entity_fil_asr_f1_micro": 0.0,
                    "entity_PER_precision": -1.0,
                    "entity_PER_recall": 0.0,
                    "entity_PER_f1": -1.0,
                    "entity_ORG_precision": 0.0,
                    "entity_ORG_recall": 0.0,
                    "entity_ORG_f1": 0.0,
                    "token_all_loss": 2.0,
                    "token_all_acc": 0.75,
                    "token_all_precision_micro": 0.75,
                    "token_all_precision_macro": 0.50,
                    "token_all_recall_micro": 0.75,
                    "token_all_recall_macro": 0.66,
                    "token_all_f1_micro": 0.75,
                    "token_all_f1_macro": 0.55,
                    "token_fil_numberofclasses_macro": 1,
                    "token_fil_precision_micro": 0.5,
                    "token_fil_precision_macro": 0.25,
                    "token_fil_recall_micro": 0.5,
                    "token_fil_recall_macro": 0.5,
                    "token_fil_f1_micro": 0.5,
                    "token_fil_f1_macro": 0.33,
                    "token_O_precision": 1.0,
                    "token_O_recall": 1.0,
                    "token_O_f1": 1.0,
                    "token_PER_precision": -1.0,
                    "token_PER_recall": 0.0,
                    "token_PER_f1": -1.0,
                    "token_ORG_precision": 0.5,
                    "token_ORG_recall": 1.0,
                    "token_ORG_f1": 0.66,
                },
                2.0,
            ),
        ],
    )
    def test_execute(
        self,
        phase: str,
        outputs: List[torch.Tensor],
        epoch_metrics: Dict[str, float],
        epoch_loss: float,
    ) -> None:
        test_epoch_metrics, _, _, test_epoch_loss = self.ner_model_evaluation.execute(
            phase, outputs
        )

        a = set(list(test_epoch_metrics.keys()))
        b = set(list(epoch_metrics.keys()))
        assert len(a.difference(b)) == 0, a.difference(b)
        assert len(b.difference(a)) == 0, b.difference(a)

        for k in list(
            set(list(test_epoch_metrics.keys()) + list(epoch_metrics.keys()))
        ):
            assert test_epoch_metrics[k] == pytest_approx(
                epoch_metrics[k]
            ), f"key = {k}: test_epoch_metrics = {test_epoch_metrics[k]} != {epoch_metrics[k]}"



import pytest
from typing import List, Optional
from nerblackbox.modules.ner_training.annotation_tags.annotation import Annotation


########################################################################################################################
########################################################################################################################
########################################################################################################################
class TestAnnotation:

    ####################################################################################################################
    @pytest.mark.parametrize(
        "classes, " "processed_classes",
        [
            (
                ["O", "PER", "ORG", "MISC"],
                ["O", "MISC", "ORG", "PER"],
            ),
            (
                ["O", "B-PER", "B-ORG", "B-MISC"],
                [
                    "O",
                    "B-MISC",
                    "B-ORG",
                    "B-PER",
                    "I-MISC",
                    "I-ORG",
                    "I-PER",
                ],
            ),
            (
                ["O", "B-PER", "B-ORG", "B-MISC", "L-MISC"],
                [
                    "O",
                    "B-MISC",
                    "B-ORG",
                    "B-PER",
                    "I-MISC",
                    "I-ORG",
                    "I-PER",
                    "L-MISC",
                    "L-ORG",
                    "L-PER",
                    "U-MISC",
                    "U-ORG",
                    "U-PER",
                ],
            ),
        ],
    )
    def test_ensure_completeness_in_case_of_bio_or_bilou_tags(
        self,
        classes: List[str],
        processed_classes: List[str],
    ) -> None:
        annotation = Annotation(classes)
        assert (
            annotation.classes == processed_classes
        ), f"annotation.classes = {annotation.classes} != {processed_classes}"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "classes, " "processed_classes",
        [
            (
                ["O", "PER", "ORG", "MISC"],
                ["O", "MISC", "ORG", "PER"],
            ),
            (
                ["PER", "ORG", "O", "MISC"],
                ["O", "MISC", "ORG", "PER"],
            ),
            (
                ["O", "B-PER", "I-MISC", "B-ORG", "I-PER", "B-MISC", "I-ORG"],
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
            ),
        ],
    )
    def test_sort(
        self,
        classes: List[str],
        processed_classes: List[str],
    ) -> None:
        annotation = Annotation(classes)
        assert (
            annotation.classes == processed_classes
        ), f"annotation.classes = {annotation.classes} != {processed_classes}"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "classes, " "annotation_scheme",
        [
            (
                ["O", "MISC", "ORG", "PER"],
                "plain",
            ),
            (
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
                "bio",
            ),
            (
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER", "L-MISC"],
                "bilou",
            ),
        ],
    )
    def test_determine_annotation_scheme(
        self,
        classes: List[str],
        annotation_scheme: str,
    ) -> None:
        annotation = Annotation(classes)
        assert (
            annotation.scheme == annotation_scheme
        ), f"ERROR! annotation.scheme = {annotation.scheme} != {annotation_scheme}"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "new_scheme, " "old_classes, " "new_classes",
        [
            # 1. plain -> plain
            (
                "plain",
                ["O", "MISC", "ORG", "PER"],
                ["O", "MISC", "ORG", "PER"],
            ),
            # 2. plain -> bio
            (
                "bio",
                ["O", "MISC", "ORG", "PER"],
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
            ),
            # 3. plain -> bilou
            (
                "bilou",
                ["O", "MISC", "ORG", "PER"],
                [
                    "O",
                    "B-MISC",
                    "B-ORG",
                    "B-PER",
                    "I-MISC",
                    "I-ORG",
                    "I-PER",
                    "L-MISC",
                    "L-ORG",
                    "L-PER",
                    "U-MISC",
                    "U-ORG",
                    "U-PER",
                ],
            ),
            # 4. bio -> plain
            (
                "plain",
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
                ["O", "MISC", "ORG", "PER"],
            ),
            # 5. bio -> bio
            (
                "bio",
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
            ),
            # 6. bio -> bilou
            (
                "bilou",
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
                [
                    "O",
                    "B-MISC",
                    "B-ORG",
                    "B-PER",
                    "I-MISC",
                    "I-ORG",
                    "I-PER",
                    "L-MISC",
                    "L-ORG",
                    "L-PER",
                    "U-MISC",
                    "U-ORG",
                    "U-PER",
                ],
            ),
            # 7. bilou -> plain
            (
                "plain",
                [
                    "O",
                    "B-MISC",
                    "B-ORG",
                    "B-PER",
                    "I-MISC",
                    "I-ORG",
                    "I-PER",
                    "L-MISC",
                    "L-ORG",
                    "L-PER",
                    "U-MISC",
                    "U-ORG",
                    "U-PER",
                ],
                ["O", "MISC", "ORG", "PER"],
            ),
            # 8. bilou -> bio
            (
                "bio",
                [
                    "O",
                    "B-MISC",
                    "B-ORG",
                    "B-PER",
                    "I-MISC",
                    "I-ORG",
                    "I-PER",
                    "L-MISC",
                    "L-ORG",
                    "L-PER",
                    "U-MISC",
                    "U-ORG",
                    "U-PER",
                ],
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
            ),
            # 9. bilou -> bilou
            (
                "bilou",
                [
                    "O",
                    "B-MISC",
                    "B-ORG",
                    "B-PER",
                    "I-MISC",
                    "I-ORG",
                    "I-PER",
                    "L-MISC",
                    "L-ORG",
                    "L-PER",
                    "U-MISC",
                    "U-ORG",
                    "U-PER",
                ],
                [
                    "O",
                    "B-MISC",
                    "B-ORG",
                    "B-PER",
                    "I-MISC",
                    "I-ORG",
                    "I-PER",
                    "L-MISC",
                    "L-ORG",
                    "L-PER",
                    "U-MISC",
                    "U-ORG",
                    "U-PER",
                ],
            ),
            # 10. unknown
            (
                "xyz",
                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
                None,
            ),
        ],
    )
    def test_change_scheme(
        self,
        new_scheme: str,
        old_classes: List[str],
        new_classes: Optional[List[str]],
    ) -> None:
        if new_classes is None:
            with pytest.raises(Exception):
                _ = Annotation(old_classes).change_scheme(new_scheme=new_scheme)
        else:
            annotation = Annotation(old_classes).change_scheme(new_scheme=new_scheme)
            assert (
                annotation.scheme == new_scheme
            ), f"ERROR! annotation.scheme = {annotation.scheme} != {new_scheme}"
            assert (
                annotation.classes == new_classes
            ), f"annotation.classes = {annotation.classes} != {new_classes}"



import pytest
from typing import List, Dict, Tuple, Union
from nerblackbox.modules.ner_training.data_preprocessing.text_encoder import TextEncoder

Encodings = List[Tuple[int, str, str]]
Predictions = List[Dict[str, Union[str, Dict]]]


########################################################################################################################
########################################################################################################################
########################################################################################################################
class TestTextEncoder:

    ####################################################################################################################
    @pytest.mark.parametrize(
        "encodings_mapping, text_list, predictions_list, predictions_encoded_list",
        [
            # Example 1: level = entity
            (
                {"\n": "[NEWLINE]", "\t": "[TAB]"},
                ["an\n example"],
                [
                    [
                        {
                            "char_start": "4",
                            "char_end": "11",
                            "token": "example",
                            "tag": "TAG",
                        }
                    ]
                ],
                [
                    [
                        {
                            "char_start": "12",
                            "char_end": "19",
                            "token": "example",
                            "tag": "TAG",
                        }
                    ]
                ],
            ),
            # Example 2: level = word
            (
                {"\n": "[NEWLINE]", "\t": "[TAB]"},
                ["an\n example"],
                [
                    [
                        {"char_start": "0", "char_end": "2", "token": "an", "tag": "O"},
                        {"char_start": "2", "char_end": "3", "token": "\n", "tag": "O"},
                        {
                            "char_start": "4",
                            "char_end": "11",
                            "token": "example",
                            "tag": "TAG",
                        },
                    ]
                ],
                [
                    [
                        {"char_start": "0", "char_end": "2", "token": "an", "tag": "O"},
                        {
                            "char_start": "2",
                            "char_end": "11",
                            "token": "[NEWLINE]",
                            "tag": "O",
                        },
                        {
                            "char_start": "12",
                            "char_end": "19",
                            "token": "example",
                            "tag": "TAG",
                        },
                    ]
                ],
            ),
        ],
    )
    def tests(
        self,
        encodings_mapping: Dict[str, str],
        text_list: List[str],
        predictions_list: List[Predictions],
        predictions_encoded_list: List[Predictions],
    ) -> None:
        text_encoder = TextEncoder(encodings_mapping)

        text_encoded_list, encodings_list = text_encoder.encode(text_list)
        test_text_list, test_predictions_list = text_encoder.decode(
            text_encoded_list, encodings_list, predictions_encoded_list
        )
        assert (
            test_text_list == text_list
        ), f"ERROR! test_text_list = {test_text_list} != text_list = {text_list}"
        assert (
            test_predictions_list == predictions_list
        ), f"ERROR! test_predictions_list = {test_predictions_list} != predictions_list = {predictions_list}"



from typing import List, Optional
from itertools import product


class LoggedMetrics:
    # metric,                 str: "loss", "acc", "precision", "recall", "f1"
    # phases,           List[str]: subset of ["train", "val", "test"]
    # levels,           List[str]: subset of ["token", "entity"]
    # tag_groups,       List[str]: subset of ["all", "fil", "ind"]
    # averaging_groups: List[str]: subset of ["simple", "micro", "macro"]
    logged_metrics = [
        # GENERAL
        ("loss", ["train", "val", "test"], ["all"], ["token"], ["simple"]),
        ("acc", ["train", "val", "test"], ["all"], ["token"], ["simple"]),
        # VAL
        ("f1", ["val"], ["all"], ["token"], ["micro"]),
        ("f1", ["val"], ["fil"], ["token", "entity"], ["micro"]),
        # TEST
        ("precision", ["test"], ["all"], ["token"], ["micro", "macro"]),
        ("precision", ["test"], ["fil"], ["token", "entity"], ["micro", "macro"]),
        ("precision", ["test"], ["ind"], ["token", "entity"], ["micro"]),
        ("precision", ["test"], ["O"], ["token"], ["micro"]),
        ("recall", ["test"], ["all"], ["token"], ["micro", "macro"]),
        ("recall", ["test"], ["fil"], ["token", "entity"], ["micro", "macro"]),
        ("recall", ["test"], ["ind"], ["token", "entity"], ["micro"]),
        ("recall", ["test"], ["O"], ["token"], ["micro"]),
        ("f1", ["test"], ["all"], ["token"], ["micro", "macro"]),
        ("f1", ["test"], ["fil"], ["token", "entity"], ["micro", "macro"]),
        ("f1", ["test"], ["ind"], ["token", "entity"], ["micro"]),
        ("f1", ["test"], ["O"], ["token"], ["micro"]),
        # special:
        ("numberofclasses", ["test"], ["fil"], ["token", "entity"], ["macro"]),
        ("asr_precision", ["test"], ["fil"], ["entity"], ["micro"]),
        ("asr_recall", ["test"], ["fil"], ["entity"], ["micro"]),
        ("asr_f1", ["test"], ["fil"], ["entity"], ["micro"]),
        ("asr_abidance", ["test"], ["fil"], ["entity"], ["simple"]),
    ]

    @classmethod
    def as_flat_list(cls):
        """
        return logged_metrics as flat list
        ----------------------------------
        :return: [list] of [str], e.g. ['token_all_precision_micro', 'token_fil_precision_macro', ..]
        """
        flat_list = list()
        for metric, _, tag_groups, levels, averaging_groups in cls.logged_metrics:
            for tag_group, level in product(tag_groups, levels):
                if averaging_groups == ["simple"]:
                    flat_list.append(f"{level}_{tag_group}_{metric}")
                else:
                    for averaging_group in averaging_groups:
                        flat_list.append(
                            f"{level}_{tag_group}_{metric}_{averaging_group}"
                        )
        return flat_list

    def __init__(self):
        for metric, phases, tag_groups, levels, averaging_groups in self.logged_metrics:
            assert isinstance(metric, str)
            assert isinstance(phases, list)
            assert isinstance(tag_groups, list)
            assert isinstance(levels, list)
            assert isinstance(averaging_groups, list)

    def get_metrics(
        self,
        required_tag_groups: Optional[List[str]] = None,
        required_phases: Optional[List[str]] = None,
        required_levels: Optional[List[str]] = None,
        required_averaging_groups: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ) -> List[str]:
        """
        get metrics, filtered
        ---------------------
        :param required_tag_groups:       [list] of tag_groups that are required, e.g. ['all']
        :param required_phases:           [list] of phases that are required, e.g. ['train']
        :param required_levels:           [list] of levels that are required, e.g. ['token']
        :param required_averaging_groups: [list] averaging groups that are required, e.g. ['micro', 'macro']
        :param exclude:                   [list] of metrics to exclude, e.g. ['loss']
        :return: filtered_metrics:        [list] of [str], e.g. ['precision', 'recall']
        """
        filtered_metrics = list()
        for metric, phases, tag_groups, levels, averaging_groups in self.logged_metrics:
            add_metric_1 = (
                True
                if required_tag_groups is None
                else all(
                    [
                        required_tag_group in tag_groups
                        for required_tag_group in required_tag_groups
                    ]
                )
            )
            add_metric_2 = (
                True
                if required_phases is None
                else all(
                    [required_phase in phases for required_phase in required_phases]
                )
            )
            add_metric_3 = (
                True
                if required_levels is None
                else all(
                    [required_level in levels for required_level in required_levels]
                )
            )
            add_metric_4 = (
                True
                if required_averaging_groups is None
                else all(
                    [
                        required_averaging_group in averaging_groups
                        for required_averaging_group in required_averaging_groups
                    ]
                )
            )
            add_metric_5 = True if exclude is None else metric not in exclude

            if (
                add_metric_1
                and add_metric_2
                and add_metric_3
                and add_metric_4
                and add_metric_5
            ):
                filtered_metrics.append(metric)

        return list(set(filtered_metrics))



import pytest
from typing import List, Dict
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example import (
    InputExample,
)
from nerblackbox.modules.ner_training.annotation_tags.input_examples_utils import (
    InputExamplesUtils,
)
from nerblackbox.modules.ner_training.annotation_tags.annotation import Annotation


########################################################################################################################
########################################################################################################################
########################################################################################################################
class TestInputExamplesUtils:

    ####################################################################################################################
    @pytest.mark.parametrize(
        "tag_classes, " "input_examples, " "tag_classes_bio, " "input_examples_bio",
        [
            (
                ["O", "PER", "ORG", "MISC"],
                {
                    "train": [
                        InputExample(
                            guid="",
                            text="På skidspår.se kan längdskidåkare själva betygsätta förhållandena i spåren .",
                            tags="O MISC O O O O O O O O",
                        ),
                    ],
                    "val": [
                        InputExample(
                            guid="",
                            text="Fastigheten är ett landmärke designad av arkitekten Robert Stern .",
                            tags="O O O O O O O PER PER O",
                        ),
                    ],
                    "test": [
                        InputExample(
                            guid="",
                            text="Apple noteras för 85,62 poäng , vilket är den högsta siffran någonsin i undersökningen .",
                            tags="ORG O O O O O O O O O O O O O O",
                        ),
                    ],
                    "predict": [
                        InputExample(
                            guid="",
                            text="På skidspår.se kan längdskidåkare själva betygsätta förhållandena i spåren .",
                            tags="O MISC O O O O O O O O",
                        ),
                        InputExample(
                            guid="",
                            text="Fastigheten är ett landmärke designad av arkitekten Robert Stern .",
                            tags="O O O O O O O PER PER O",
                        ),
                        InputExample(
                            guid="",
                            text="Apple noteras för 85,62 poäng , vilket är den högsta siffran någonsin i undersökningen .",
                            tags="ORG O O O O O O O O O O O O O O",
                        ),
                    ],
                },
                ["O", "B-PER", "B-ORG", "B-MISC", "I-PER", "I-ORG", "I-MISC"],
                {
                    "train": [
                        InputExample(
                            guid="",
                            text="På skidspår.se kan längdskidåkare själva betygsätta förhållandena i spåren .",
                            tags="O B-MISC O O O O O O O O",
                        ),
                    ],
                    "val": [
                        InputExample(
                            guid="",
                            text="Fastigheten är ett landmärke designad av arkitekten Robert Stern .",
                            tags="O O O O O O O B-PER I-PER O",
                        ),
                    ],
                    "test": [
                        InputExample(
                            guid="",
                            text="Apple noteras för 85,62 poäng , vilket är den högsta siffran någonsin i undersökningen .",
                            tags="B-ORG O O O O O O O O O O O O O O",
                        ),
                    ],
                },
            ),
        ],
    )
    def tests(
        self,
        tag_classes: List[str],
        input_examples: Dict[str, List[InputExample]],
        tag_classes_bio: List[str],
        input_examples_bio: Dict[str, List[InputExample]],
    ) -> None:
        def _test(
            _test_input_examples: Dict[str, List[InputExample]],
            _input_examples: Dict[str, List[InputExample]],
            _test_name: str,
        ):
            """
            test _test_input_examples against ground truth _input_examples

            Args:
                _test_input_examples: input examples to be tested
                _input_examples:      input examples to be tested against
                _test_name:           _test_input_examples as string
            """
            for phase in ["train", "val", "test"]:
                assert (
                    len(_test_input_examples[phase]) == 1
                ), f"ERROR! len({_test_name}[{phase}]) = {len(_test_input_examples[phase])} should be 1."
                assert (
                    _test_input_examples[phase][0].text
                    == _input_examples[phase][0].text
                ), f"phase = {phase}: {_test_name}.text = {_test_input_examples[phase][0].text} != {_input_examples[phase][0].text}"
                assert (
                    _test_input_examples[phase][0].tags
                    == _input_examples[phase][0].tags
                ), f"phase = {phase}: {_test_name}.tags = {_test_input_examples[phase][0].tags} != {_input_examples[phase][0].tags}"

        # b1. plain -> bio
        test_input_examples_plain2bio = InputExamplesUtils.convert_annotation_scheme(
            input_examples=input_examples,
            annotation_scheme_source=Annotation(tag_classes).scheme,
            annotation_scheme_target="bio",
        )
        _test(
            test_input_examples_plain2bio,
            input_examples_bio,
            "test_input_examples_plain2bio",
        )

        # b2. bio -> plain
        test_input_examples_bio2plain = InputExamplesUtils.convert_annotation_scheme(
            input_examples=input_examples_bio,
            annotation_scheme_source=Annotation(tag_classes_bio).scheme,
            annotation_scheme_target="plain",
        )
        _test(
            test_input_examples_bio2plain,
            input_examples,
            "test_input_examples_bio2plain",
        )

        # b3. plain -> plain
        test_input_examples_plain = InputExamplesUtils.convert_annotation_scheme(
            input_examples=input_examples,
            annotation_scheme_source=Annotation(tag_classes).scheme,
            annotation_scheme_target="plain",
        )
        _test(test_input_examples_plain, input_examples, "test_input_examples_plain")

        # b4. bio -> bio
        test_input_examples_bio = InputExamplesUtils.convert_annotation_scheme(
            input_examples=input_examples_bio,
            annotation_scheme_source=Annotation(tag_classes_bio).scheme,
            annotation_scheme_target="bio",
        )
        _test(test_input_examples_bio, input_examples_bio, "test_input_examples_bio")

        # b5. use unknown parameter
        with pytest.raises(Exception):
            _ = InputExamplesUtils.convert_annotation_scheme(
                input_examples=input_examples,
                annotation_scheme_source=Annotation(tag_classes).scheme,
                annotation_scheme_target="xyz",
            )



from nerblackbox.modules.datasets.formatter.swedish_ner_corpus_formatter import (
    SwedishNerCorpusFormatter,
)
from nerblackbox.modules.datasets.formatter.sic_formatter import SICFormatter
from nerblackbox.modules.datasets.formatter.suc_formatter import SUCFormatter
from nerblackbox.modules.datasets.formatter.sucx_formatter import SUCXFormatter
from nerblackbox.modules.datasets.formatter.conll2003_formatter import (
    CoNLL2003Formatter,
)
from nerblackbox.modules.datasets.formatter.base_formatter import BaseFormatter
from nerblackbox.modules.datasets.formatter.swe_nerc_formatter import SweNercFormatter
from nerblackbox.modules.datasets.formatter.huggingface_datasets_formatter import (
    HuggingfaceDatasetsFormatter,
)


class AutoFormatter:
    @staticmethod
    def for_dataset(ner_dataset: str, ner_dataset_subset: str = "") -> BaseFormatter:
        """
        Args:
            ner_dataset: e.g. "conll2003"
            ner_dataset_subset: e.g. "simple_cased"

        Returns:
            formatter: e.g. CoNLL2003Formatter
        """
        if ner_dataset == "swedish_ner_corpus":
            return SwedishNerCorpusFormatter()
        elif ner_dataset == "conll2003_from_source":
            return CoNLL2003Formatter()
        elif ner_dataset == "sic":
            return SICFormatter()
        elif ner_dataset == "suc":
            return SUCFormatter()
        elif ner_dataset == "sucx":
            assert len(
                ner_dataset_subset
            ), f"ERROR! for sucx, a subset needs to be specified."
            return SUCXFormatter(ner_dataset_subset)
        elif ner_dataset == "swe_nerc":
            return SweNercFormatter()
        else:  # huggingface datasets
            existence, error_msg = HuggingfaceDatasetsFormatter.check_existence(
                ner_dataset, ner_dataset_subset
            )
            if existence:
                print(f"> ner_dataset = {ner_dataset} found in huggingface datasets")
            else:
                raise Exception(error_msg)

            compatibility, error_msg = HuggingfaceDatasetsFormatter.check_compatibility(
                ner_dataset, ner_dataset_subset
            )
            if compatibility:
                print(f"> ner_dataset = {ner_dataset} contains train/val/test splits")
            else:
                raise Exception(error_msg)

            (
                implementation,
                error_msg,
            ) = HuggingfaceDatasetsFormatter.check_implementation(
                ner_dataset, ner_dataset_subset
            )
            if implementation:
                print(f"> ner_dataset = {ner_dataset} can be parsed")
            else:
                raise Exception(error_msg)

            return HuggingfaceDatasetsFormatter(ner_dataset, ner_dataset_subset)



import json

from typing import Dict
from transformers import AutoModelForTokenClassification
from omegaconf import DictConfig
from torch.optim.lr_scheduler import LambdaLR
from torch.optim.optimizer import Optimizer

from nerblackbox.modules.ner_training.metrics.logged_metrics import LoggedMetrics
from nerblackbox.modules.ner_training.logging.mlflow_client import MLflowClient
from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger
from nerblackbox.modules.ner_training.ner_model import NerModel
from nerblackbox.modules.ner_training.annotation_tags.input_examples_utils import (
    InputExamplesUtils,
)
from nerblackbox.modules.utils.util_functions import read_encoding


class NerModelTrain(NerModel):
    def __init__(self, hparams: DictConfig):
        """
        :param hparams: attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
        """
        super().__init__(hparams)

    ####################################################################################################################
    # Abstract Base Methods ############################################################################################
    ####################################################################################################################
    def _preparations(self):
        """
        :created attr: default_logger    [DefaultLogger]
        :created attr: logged_metrics    [LoggedMetrics]
        :created attr: tokenizer         [transformers AutoTokenizer]
        :created attr: encoding          [dict] e.g. {"\n": "[NEWLINE]", "\t": "[TAB]"}
        :created attr: special_tokens    [list] of [str] e.g. ["[NEWLINE]", "[TAB]"]
        :created attr: data_preprocessor [DataPreprocessor]
        :created attr: annotation        [Annotation]
        :created attr: model             [transformers AutoModelForTokenClassification]

        :created attr: mlflow_client          [MLflowClient]
        :created attr: epoch_metrics          [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: classification_reports [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: pretrained_model_name  [str]
        :created attr: dataloader             [dict] w/ keys 'train', 'val', 'test' & values = [torch Dataloader]
        :created attr: optimizer              [torch optimizer]
        :created attr: scheduler              [torch LambdaLR]

        :return: -
        """
        # train/val/test
        self._preparations_train()  # attr: default_logger, logged_metrics, mlflow_client, ..
        self._preparations_data_general()  # attr: tokenizer, data_preprocessor
        self._preparations_data_train()  # attr: annotation, model, dataloader, optimizer, scheduler

    def _preparations_train(self):
        """
        :created attr: default_logger         [DefaultLogger]
        :created attr: logged_metrics         [LoggedMetrics]
        :created attr: mlflow_client          [MLflowClient]
        :created attr: epoch_metrics          [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: classification_reports [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: encoding               [dict] e.g. {"\n": "[NEWLINE]", "\t": "[TAB]"}
        :created attr: special_tokens         [list] of [str] e.g. ["[NEWLINE]", "[TAB]"]
        :return: -
        """
        self.default_logger = DefaultLogger(
            __file__, log_file=self.log_dirs.log_file, level=self.params.logging_level
        )

        self.logged_metrics = LoggedMetrics()

        self.mlflow_client = MLflowClient(
            experiment_name=self.params.experiment_name,
            run_name=self.params.run_name,
            log_dirs=self.log_dirs,
            logged_metrics=self.logged_metrics.as_flat_list(),
            default_logger=self.default_logger,
        )
        self.mlflow_client.log_params(
            self.params, self.hparams, experiment=self.experiment
        )

        self.epoch_metrics: Dict[str, Dict] = {"val": dict(), "test": dict()}
        self.classification_reports: Dict[str, Dict] = {"val": dict(), "test": dict()}

        self.encoding = read_encoding(self.params.dataset_name)
        self.default_logger.log_info(f"> read encoding: {self.encoding}")

        self.special_tokens = list(set(self.encoding.values()))
        self.default_logger.log_debug(
            f"> extracted special tokens: {self.special_tokens}"
        )

    def _preparations_data_train(self):
        """
        :created attr: annotation        [Annotation]
        :created attr: model             [transformers AutoModelForTokenClassification]
        :created attr: dataloader        [dict] w/ keys 'train', 'val', 'test' & values = [torch Dataloader]
        :created attr: optimizer         [torch optimizer]
        :created attr: scheduler         [torch LambdaLR]
        :return: -
        """
        # input_examples & annotation
        (
            input_examples,
            self.annotation,
        ) = self.data_preprocessor.get_input_examples_train(
            prune_ratio={
                "train": self.params.prune_ratio_train,
                "val": self.params.prune_ratio_val,
                "test": self.params.prune_ratio_test,
            },
            dataset_name=self.params.dataset_name,
            train_on_val=self.params.train_on_val,
            train_on_test=self.params.train_on_test,
        )
        self.default_logger.log_info(
            f"> annotation scheme found: {self.annotation.scheme}"
        )
        if self.params.annotation_scheme == "auto":
            self.params.annotation_scheme = self.annotation.scheme
        elif self.params.annotation_scheme != self.annotation.scheme:
            # convert annotation_classes
            input_examples = InputExamplesUtils.convert_annotation_scheme(
                input_examples=input_examples,
                annotation_scheme_source=self.annotation.scheme,
                annotation_scheme_target=self.params.annotation_scheme,
            )
            self.annotation = self.annotation.change_scheme(
                new_scheme=self.params.annotation_scheme
            )
            self.default_logger.log_info(
                f"> annotation scheme converted to {self.params.annotation_scheme}"
            )

        self.default_logger.log_debug(
            "> self.annotation.classes:", self.annotation.classes
        )

        assert isinstance(
            self.hparams, DictConfig
        ), f"ERROR! type(self.hparams) = {type(self.hparams)} should be DictConfig."
        assert isinstance(
            self._hparams, DictConfig
        ), f"ERROR! type(self._hparams) = {type(self._hparams)} should be DictConfig."

        self.hparams.annotation_classes = json.dumps(
            self.annotation.classes
        )  # save for Model
        self.hparams.encoding = json.dumps(self.encoding)  # save for Model

        # model
        self.model = AutoModelForTokenClassification.from_pretrained(
            self.pretrained_model_name,
            num_labels=len(self.annotation.classes),
            return_dict=False,
            ignore_mismatched_sizes=True,  # if pretrained model = AutoModelForTokenClassification
            # with different number of annotation classes
        )
        self.model.resize_token_embeddings(
            len(self.tokenizer)
        )  # due to additional_special_tokens

        # dataloader
        self.dataloader, _ = self.data_preprocessor.to_dataloader(
            input_examples, self.annotation.classes, batch_size=self._hparams.batch_size
        )

        # optimizer
        self.optimizer: Optimizer = self._create_optimizer(
            self._hparams.lr_max, fp16=self.params.fp16
        )

        # learning rate
        self.scheduler: LambdaLR = self._create_scheduler(
            self._hparams.lr_warmup_epochs,
            self._hparams.lr_schedule,
            self.hyperparameters.max_epochs,
            self._hparams.lr_num_cycles,
        )



import os
from typing import Tuple, List, Optional, Dict
import numpy as np
from os.path import join, isfile
import json
from argparse import Namespace
import pkg_resources
from omegaconf import OmegaConf, DictConfig

from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.utils.parameters import GENERAL, PARAMS, HPARAMS, LOG_DIRS


def get_available_datasets() -> List[str]:
    """
    get datasets that are available in DIR_DATASETS directory

    Returns:
        available datasets: e.g. ['suc', 'swedish_ner_corpus']
    """
    dir_datasets = env_variable("DIR_DATASETS")
    return [
        folder
        for folder in os.listdir(dir_datasets)
        if os.path.isdir(join(dir_datasets, folder))
    ]


def get_dataset_path(dataset: str, subset: str = "") -> str:
    """
    get dataset path for dataset

    Args:
        dataset: e.g. 'suc', 'swedish_ner_corpus'
        subset: e.g. 'original_cased'

    Returns:
        dataset_path: path to dataset directory
    """
    if len(subset):
        return join(env_variable("DIR_DATASETS"), dataset, subset)
    else:
        return join(env_variable("DIR_DATASETS"), dataset)


def read_encoding(dataset: str) -> Dict[str, str]:
    """
    if the file "encoding.json" exists for the dataset, read it and return its content
    ----------------------------------------------------------------------------------

    Args:
        dataset: e.g. 'swedish_ner_corpus'

    Returns:
        encoding: e.g. {"\n": "[NEWLINE]", "\t": "[TAB]"}
    """
    path_encoding = join(get_dataset_path(dataset), "encoding.json")
    if isfile(path_encoding):
        with open(path_encoding, "r") as file:
            encoding = json.load(file)
        assert isinstance(
            encoding, dict
        ), f"ERROR! {path_encoding} seems to contain a {type(encoding)}, should be list."
    else:
        encoding = dict()
    return encoding


def get_hardcoded_parameters(keys=False):
    """
    :param keys: [bool] whether to return [list] of keys instead of whole [dict]
    :return: _general:  [dict] w/ keys = parameter name [str] & values = type [str] --- or [list] of keys
    :return: _params:   [dict] w/ keys = parameter name [str] & values = type [str] --- or [list] of keys
    :return: _hparams:  [dict] w/ keys = parameter name [str] & values = type [str] --- or [list] of keys
    :return: _log_dirs: [dict] w/ keys = parameter name [str] & values = type [str] --- or [list] of keys
    """
    _general = {
        "experiment_name": "str",
        "run_name": "str",
        "run_name_nr": "str",
        "device": "str",
        "fp16": "bool",
        "experiment_run_name_nr": "str",
    }
    _params = {
        "dataset_name": "str",
        "annotation_scheme": "str",
        "prune_ratio_train": "float",
        "prune_ratio_val": "float",
        "prune_ratio_test": "float",
        "pretrained_model_name": "str",
        "uncased": "bool",
        "checkpoints": "bool",
        "logging_level": "str",
        "multiple_runs": "int",
        "seed": "int",
    }
    _hparams = {
        "batch_size": "int",
        "max_seq_length": "int",
        "max_epochs": "int",
        "early_stopping": "bool",
        "monitor": "str",
        "min_delta": "float",
        "patience": "int",
        "mode": "str",
        "lr_max": "float",
        "lr_schedule": "str",
        "lr_warmup_epochs": "int",
        "lr_cooldown_epochs": "int",
        "lr_cooldown_restarts": "bool",
        "lr_num_cycles": "int",
    }
    _log_dirs = {
        "mlflow": "str",
        "tensorboard": "str",
        "checkpoints": "str",
        "log_file": "str",
        "mlflow_file": "str",
    }
    if keys:
        return (
            list(_general.keys()),
            list(_params.keys()),
            list(_hparams.keys()),
            list(_log_dirs.keys()),
        )
    else:
        return _general, _params, _hparams, _log_dirs


def unify_parameters(
    _params: Namespace, _hparams: Namespace, _log_dirs: Namespace, _experiment: bool
) -> DictConfig:
    """
    unify parameters (namespaces, bool) to one namespace

    Args:
        _params:        keys = 'dataset_name', 'annotation_scheme', ..
        _hparams:       keys = 'batch_size', 'max_seq_length', ..
        _log_dirs:      keys = 'mlflow', 'tensorboard', ..
        _experiment:

    Returns:
        _lightning_hparams: keys = all keys from input namespaces + 'experiment'
    """
    _dict = dict()
    _dict.update(vars(_params))
    _dict.update(vars(_hparams))
    _dict.update(vars(_log_dirs))
    _dict.update({"experiment": _experiment})
    _lightning_hparams = Namespace(**_dict)
    _lightning_hparams.device = (
        _lightning_hparams.device.type
    )  # needs to be a string (not torch.device) for logging

    omega_conf = OmegaConf.create(vars(_lightning_hparams))
    assert (
        type(omega_conf) == DictConfig
    ), f"ERROR! type(omega_conf) = {type(omega_conf)} should be DictConfig"
    return omega_conf


def split_parameters(
    _lightning_hparams: DictConfig,
) -> Tuple[Namespace, Namespace, Namespace, bool]:
    """
    split namespace to parameters (namespaces, bool)

    Args:
        _lightning_hparams: keys = all keys from output namespaces + 'experiment'

    Returns:
        _params:            keys = 'dataset_name', 'annotation_scheme', ..
        _hparams:           keys = 'batch_size', 'max_seq_length', ..
        _log_dirs:          keys = 'mlflow', 'tensorboard', ..
        _experiment:
    """
    _lightning_hparams_dict = OmegaConf.to_container(_lightning_hparams)
    assert isinstance(
        _lightning_hparams_dict, dict
    ), f"ERROR! {type(_lightning_hparams_dict)}"

    _params = Namespace(
        **{
            str(k): v
            for k, v in _lightning_hparams_dict.items()
            if k in list(GENERAL.keys()) + list(PARAMS.keys())
        }
    )
    _hparams = Namespace(
        **{
            str(k): v
            for k, v in _lightning_hparams_dict.items()
            if k in list(HPARAMS.keys())
        }
    )
    _log_dirs = Namespace(
        **{
            str(k): v
            for k, v in _lightning_hparams_dict.items()
            if k in list(LOG_DIRS.keys())
        }
    )
    _experiment = _lightning_hparams.get("experiment")
    return _params, _hparams, _log_dirs, _experiment


def get_package_version() -> str:
    return pkg_resources.get_distribution("nerblackbox").version


def checkpoint2epoch(_checkpoint_name: str) -> int:
    """
    Args:
        _checkpoint_name: e.g. 'epoch=2.ckpt' or 'epoch=2_v0.ckpt'

    Returns:
        _epoch:           e.g. 2
    """
    return int(_checkpoint_name.split("epoch=")[-1].split("_")[0].replace(".ckpt", ""))


def epoch2checkpoint(_epoch: int) -> str:
    """
    Args:
        _epoch:           e.g. 2

    Returns:
        _checkpoint_name: e.g. 'epoch=2.ckpt'
    """
    return f"epoch={_epoch}.ckpt"


def get_run_name(_run_name_nr: str) -> str:
    """
    Args:
        _run_name_nr: e.g. 'runA-1'

    Returns:
        _run_name:    e.g. 'runA'
    """
    return _run_name_nr.split("-")[0]


def get_run_name_nr(_run_name: str, _run_nr: int) -> str:
    """
    Args:
        _run_name: e.g. 'runA'
        _run_nr:   e.g. 1

    Returns:
        _run_name_nr: e.g. 'runA-1'
    """
    return f"{_run_name}-{_run_nr}"


def compute_mean_and_dmean(values: List[float]) -> Tuple[float, Optional[float]]:
    """
    compute mean and its error dmean = std deviation / sqrt(N)

    Args:
        values: e.g. [1., 2.]

    Returns:
        mean:   e.g. 1.5
        dmean:  e.g. 0.35355 = 0.5 / sqrt(2)
    """
    if len(values) == 0:
        return -1, None
    elif len(values) == 1:
        return values[0], None
    else:  # i.e. if len(values) > 1:
        return float(np.mean(values)), float(np.std(values) / np.sqrt(len(values)))
